# coding: utf-8

class FileFormatError(Exception):
    pass

from .ase_Atoms import load as load_ase_Atoms
from .cif import load as load_cif
from .pymatgen_Structure import load as load_pymatgen_Structure
from .table import load as load_table
from .system_model import load as load_system_model
from .prototype import load as load_prototype
from .crystal import load as load_crystal
from .poscar import load as load_poscar
from .atom_data import load as load_atom_data
from .atom_dump import load as load_atom_dump
from .spglib_cell import load as load_spglib_cell
from .phonopy_Atoms import load as load_phonopy_Atoms

__all__ = ['FileFormatError', 'load', 'load_ase_Atoms', 'load_pymatgen_Structure',
           'load_table', 'load_system_model', 'load_poscar', 'load_atom_data',
           'load_atom_dump', 'load_cif', 'load_spglib_cell', 'load_phonopy_Atoms',
           'load_prototype', 'load_crystal'
           ]

def load(style, *args, **kwargs):
    """
    Load a System.
    
    Parameters
    ----------
    style : str
        Indicates the format of the content to load as an atomman.System
    args 
        Any positional-dependent arguments to pass to the underlying load methods.
    kwargs
        Any keyword arguments to pass to the underlying load methods.
        
    Returns
    -------
    system : atomman.System
        The system object associated with the data model.
    """
    
    if style == 'system_model':
        return load_system_model(*args, **kwargs)
    
    elif style == 'prototype':
        return load_prototype(*args, **kwargs)

    elif style == 'crystal':
        return load_crystal(*args, **kwargs)

    elif style == 'cif':
        return load_cif(*args, **kwargs)
    
    elif style == 'atom_data':
        return load_atom_data(*args, **kwargs)
    
    elif style == 'atom_dump':
        return load_atom_dump(*args, **kwargs)
    
    elif style == 'table':
        return load_table(*args, **kwargs)
    
    elif style == 'ase_Atoms':
        return load_ase_Atoms(*args, **kwargs)
    
    elif style == 'phonopy_Atoms':
        return load_phonopy_Atoms(*args, **kwargs)
    
    elif style == 'pymatgen_Structure':
        return load_pymatgen_Structure(*args, **kwargs)
    
    elif style == 'poscar':
        return load_poscar(*args, **kwargs)
    
    elif style == 'spglib_cell':
        return load_spglib_cell(*args, **kwargs)
    
    else:
        raise ValueError('Unsupported style')


# coding: utf-8

# http://www.numpy.org/
import numpy as np

class Plane():
    """
    Class representing a plane in space. Useful for making slices.
    """
    def __init__(self, normal, point):
        """
        Defines a plane in space.
        
        Parameters
        ----------
        normal : array-like object
            3D normal vector of the plane.
        point : array-like object
            3D vector coordinate of any point in the plane.
        """
        
        # Set values
        self.normal = normal 
        self.point = point
    
    @property
    def normal(self):
        """numpy.NDArray : 3D normal unit vector of the plane."""
        return self.__normal
    
    @normal.setter
    def normal(self, value):
        
        # Check that value is proper
        value = np.asarray(value)
        assert value.shape == (3,), "point must be a 3D vector"
        
        # Save normalvector as a unit vector
        self.__normal = value / np.linalg.norm(value)
    
    @property
    def point(self):
        """numpy.NDArray : a 3D vector position on the plane."""
        return self.__point
    
    @point.setter
    def point(self, value):
        
        # Check that value is proper
        value = np.asarray(value)
        assert value.shape == (3,), "point must be a 3D vector"
        
        # Save
        self.__point = value
    
    def below(self, pos, inclusive=True):
        """
        Indicates if position(s) are below the plane.  Note that identifying
        points as above or below is dependent on the sign of the plane normal.
        
        Parameters
        ----------
        pos : array-like object
            Nx3 array of coordinates. 
        inclusive : bool, optional
            Indicates if points in the plane are to be included.
            Default value is True.
            
        Returns
        -------
        numpy.NDArray
            N array of bool values: True if below the plane
        """
        
        # Ensure pos is a numpy array 
        pos = np.asarray(pos)
        
        # Determine above/below based on dot product with plane normal vector
        normpoint = np.dot(self.normal, self.point)
        normpos = np.inner(self.normal, pos)
        
        if inclusive:
            return normpos <= normpoint
        else:
            return normpos < normpoint
    
    def above(self, pos, inclusive=False):
        """
        Indicates if position(s) are above the plane.  Note that identifying
        points as above or below is dependent on the sign of the plane normal.
        
        Parameters
        ----------
        pos : array-like object
            Nx3 array of coordinates. 
        inclusive : bool, optional
            Indicates if points in the plane are to be included.
            Default value is False.
        
        Returns
        -------
        numpy.NDArray
            N array of bool values: True if above the plane
        """
        return ~self.below(pos, inclusive=not inclusive)


# coding: utf-8
# Standard Python libraries
from collections import OrderedDict

# http://www.numpy.org/
import numpy as np

# https://github.com/usnistgov/DataModelDict
from DataModelDict import DataModelDict as DM

# atomman imports
import atomman.unitconvert as uc
from ... import Atoms, Box, System

def load(model, symbols=None, key='atomic-system', index=0):
    """
    Read in a data model containing a crystal structure.
    
    Parameters
    ----------
    model : str, file-like object or DataModelDict
        The data model to read.
    symbols : tuple, optional
        Allows the list of element symbols to be assigned during loading.
    key : str, optional
        The key identifying the root element for the system definition.
        Default value is 'atomic-system'.
    index : int, optional.
        If the full model has multiple key entries, the index specifies which
        to access.  Default value is 0 (first, or only entry).
    
    Returns
    -------
    system : atomman.System
        The system object associated with the data model.
    """
    
    # Pull system model out of data model using key and index
    a_sys = DM(model).finds(key)
    if len(a_sys) == 0:
        raise KeyError(key + ' not found in model')
    try:
        a_sys = a_sys[index]
    except:
        raise IndexError('Invalid index ' + str(index) + ' for model key ' + key)
    
    # Load new style model
    if 'box' in a_sys:
        system = System(model=DM([('atomic-system', a_sys)]), symbols=symbols)

    # Load old style model
    else:
        # Extract crystal system and box values
        c_system = list(a_sys['cell'].keys())[0]
        cell = a_sys['cell'][c_system]
        
        if c_system == 'cubic': 
            a = b = c = uc.value_unit(cell['a'])
            alpha = beta = gamma = 90.0
        
        elif c_system == 'hexagonal':
            a = b = uc.value_unit(cell['a'])
            c = uc.value_unit(cell['c'])
            alpha = beta = 90.0
            gamma = 120.0
        
        elif c_system == 'tetragonal':
            a = b = uc.value_unit(cell['a'])
            c = uc.value_unit(cell['c'])
            alpha = beta = gamma = 90.0
        
        elif c_system == 'trigonal' or c_system == 'rhombohedral':
            a = b = c = uc.value_unit(cell['a'])
            alpha = beta = gamma = cell['alpha']
        
        elif c_system == 'orthorhombic':
            a = uc.value_unit(cell['a'])
            b = uc.value_unit(cell['b'])
            c = uc.value_unit(cell['c'])
            alpha = beta = gamma = 90.0
        
        elif c_system == 'monoclinic':
            a = uc.value_unit(cell['a'])
            b = uc.value_unit(cell['b'])
            c = uc.value_unit(cell['c'])
            alpha = gamma = 90.0
            beta = cell['beta']
        
        elif c_system == 'triclinic':
            a = uc.value_unit(cell['a'])
            b = uc.value_unit(cell['b'])
            c = uc.value_unit(cell['c'])
            alpha = cell['alpha']
            beta = cell['beta']
            gamma = cell['gamma']
        
        box = Box(a=a, b=b, c=c, alpha=alpha, beta=beta, gamma=gamma)
        
        # Count atypes and generate list of symbols if given
        atoms = []
        scale = None
        
        all_atypes = np.array(a_sys.finds('component'))
        all_symbols = np.array(a_sys.finds('symbol'))
        all_elements = np.array(a_sys.finds('element'))
        
        if len(all_atypes) == 0:
            if len(all_symbols) != 0:
                lsymbols, atypes = np.unique(all_symbols, return_inverse=True)
            elif len(all_elements) != 0:
                lsymbols, atypes = np.unique(all_elements, return_inverse=True)
            else:
                raise ValueError('No atom components, symbols or elements listed')
        
        else:
            atypes = all_atypes
            lsymbols = [None for i in range(max(all_atypes))]
            
            if len(all_elements) != 0 and len(all_symbols) == 0:
                all_symbols = all_elements
            
            if len(all_symbols) != 0:
                assert len(all_symbols) == len(atypes)
                sym_dict = {}
                for atype, symbol in zip(atypes, all_symbols):
                    if atype not in sym_dict:
                        sym_dict[atype] = symbol
                    else:
                        assert sym_dict[atype] == symbol
                
                for atype, symbol in sym_dict.items():
                    lsymbols[atype-1] = symbol
        
        # Use lsymbols if symbols parameter is not given.
        if symbols is None:
            symbols = lsymbols
        
        # Read per-atom properties
        natoms = len(atypes)
        prop = OrderedDict()
        prop['atype'] = atypes
        prop['pos'] = np.zeros((natoms, 3), dtype='float64')
        count = 0
        
        pos_units = []
        for atom in a_sys.iteraslist('atom'):
            
            # Read in pos for atom and unit info
            prop['pos'][count] = uc.value_unit(atom['position'])
            pos_units.append(atom['position'].get('unit', None))
            
            # Add other per-atom properties
            for property in atom.iteraslist('property'):
                if property['name'] not in prop:
                    value = uc.value_unit(property)
                    shape = (natoms, ) + value.shape
                    prop[property['name']] = np.zeros(shape, dtype=value.dtype)
                    
                prop[property['name']][count] = uc.value_unit(property)
            count += 1
        
        pos_unit = np.unique(pos_units)
        assert len(pos_unit) == 1, 'Mixed units for positions'
        if pos_unit[0] == 'scaled':
            scale=True
        else:
            scale=False
        
        atoms = Atoms(**prop)
        system = System(box=box, atoms=atoms, scale=scale, symbols=symbols)
    
    return system


# coding: utf-8

# http://www.numpy.org/
import numpy as np

def compositionstr(symbols, counts):
    """
    Generates a reduced composition string based on symbols and counts.
    
    Parameters
    ----------
    symbols : list
        The model symbol for a site.
    count : list
        How many sites are occupied by each symbol.
    
    Returns
    -------
    str
        The reduced composition string.
    """
    assert len(symbols) == len(counts), 'Symbols and counts must be the same length'
    
    sym_dict = {}
    for symbol, count in zip(symbols, counts):
        if symbol in sym_dict:
            sym_dict[symbol] += count
        else:
            sym_dict[symbol] = count
    
    gcd = np.gcd.reduce(list(sym_dict.values())) # pylint: disable=no-member
    
    composition =''
    for symbol in sorted(sym_dict):
        count = sym_dict[symbol] // gcd
        if sym_dict[symbol] > 0:
            composition += symbol
            if count != 1:
                composition += str(count)
    
    return composition


# coding: utf-8
# Standard Python libraries
from copy import deepcopy
from collections import OrderedDict

# http://www.numpy.org/
import numpy as np

# https://pandas.pydata.org/
import pandas as pd

# https://github.com/usnistgov/DataModelDict
from DataModelDict import DataModelDict as DM

# atomman imports
import atomman.unitconvert as uc
from ..tools import indexstr

class Atoms(object):
    """
    Class for representing a collection of atoms.
    """
    
    class PropertyDict(OrderedDict):
        """Extends OrderedDict to work with Atoms"""
        
        def __init__(self, host):
            """PropertyDict needs to know its host Atoms object."""
            self.__host = host
            super(Atoms.PropertyDict, self).__init__()
        
        def __setitem__(self, key, value):
            """
            Modifies OrderedDict__setitem__() such that
            1. All values are converted to numpy.ndarrays.
            2. For new keys, checks are done that data is compatible with host.natoms.
            3. New keys are also added as attributes to host Atoms object, if allowed.
            4. For existing keys, new values are saved over the old ones as opposed to
            only changing the name assignment.

            Parameters
            ----------
            key : str
                The property key name to assign values to.
            value : numpy.ndarray
                The values to assign.
            """
            # Shortcut to host
            host = self.__host
            
            # Python 2: change to unicode if needed.
            try:
                key = key.decode('UTF-8')
            except:
                pass
            
            # Convert to numpy.ndarray if needed
            value = np.asarray(value)
            
            # Broadcast if needed and allowed
            if value.shape == ():
                value = np.array(np.broadcast_to(value, (host.natoms,) + value.shape))
            elif value.shape[0] == 1:
                value = np.array(np.broadcast_to(value, (host.natoms,) + value.shape[1:]))
            elif value.shape[0] != host.natoms:
                raise ValueError('First dimension of value must be 1 or natoms')
            
            # Check that atype values are 1 or greater
            if key == 'atype' and len(value) > 0 and np.min(value) < 1:
                raise ValueError('atype values must be >= 1')

            # If key is already assigned, save value over existing values
            if key in self.keys():
                self[key][:] = value
            
            # Otherwise, set new item and try assigning attribute to host
            else:
                super(Atoms.PropertyDict, self).__setitem__(key, value)
                try:
                    assert key not in dir(host)
                    super(Atoms, host).__setattr__(key, value) # pylint: disable=bad-super-call
                except:
                    pass
    
    def __init__(self, natoms=None, atype=None, pos=None, prop=None, model=None, 
                 safecopy=False, **kwargs):
        """
        Class initializer.
        
        Parameters
        ==========
        natoms : int, optional
            The number of atoms.  If not given, will be inferred from other
            parameters if possible, or set to 1 if not possible.
        atype : int or list/ndarray of int, optional
            The integer atomic types to assign to all atoms.  Default is to
            set all atypes to 1.
        pos : list/ndarray of float, optional
            The atomic positions to assign to all atoms.  Default is to set
            each atom's position to [0,0,0].
        model : str or DataModelDict, optional
            File path or content of a JSON/XML data model containing all
            atom information.  Cannot be given with any other parameters.
        prop : dict, optional
            Dictionary containing all per-atom properties.  Can be used
            instead of atype, pos, and kwargs.  This is for backwards
            compatibility support with atomman version 1.
        safecopy : bool, optional
            Flag indicating if values are to be copied before setting.  For
            property values given as numpy arrays, direct setting (False,
            default) may result in the Atoms' property pointing to the original
            numpy array.  Using safecopy=True deep copies the property values
            before setting to avoid this.  Note that safecopy=True may be
            considerably slower for large numbers of atoms and/or properties.
        kwargs : any
            All additional key/value pairs are assigned as properties.
        
        Returns
        =======
        Atoms
            The Atoms object.
        """
        
        # Check for model
        if model is not None:
            try:
                assert natoms is None
                assert atype is None
                assert pos is None
                assert prop is None
                assert len(kwargs) == 0
            except:
                raise ValueError('model cannot be given with any other parameters')
            
            # Extract natoms and properties from data model
            model = DM(model).find('atoms')
            natoms = model['natoms']
            prop = OrderedDict()
            for propmodel in model.aslist('property'):
                prop[propmodel['name']] = uc.value_unit(propmodel['data'])
        
        # Check for prop dictionary
        if prop is not None:
            if atype is not None and pos is not None and len(kwargs) > 0:
                raise ValueError('prop dict cannot be given with keyword properties')
            
            # Divide prop into keyword arguments
            atype = prop.pop('atype', None)
            pos = prop.pop('pos', None)
            kwargs = prop
        
        # Check atype parameter values
        if atype is not None:
            atype = np.asarray(atype, dtype='uint64')
            
            # Handle single atype
            if atype.ndim == 0:
                natoms_atype = 1
            
            # Handle list of atype
            elif atype.ndim == 1:
                natoms_atype = atype.shape[0]
            else:
                raise ValueError('Only one atype per atom allowed')
        else:
            atype = np.array([1], dtype='uint64')
            natoms_atype = 1
        
        # Check pos parameter values
        if pos is not None:
            pos = np.asarray(pos, dtype='float64')
            
            # Handle single pos
            if pos.ndim == 1:
                if pos.shape[0] == 3:
                    natoms_pos = 1
                else:
                    raise ValueError('pos per atom must be 3-dimensional')
            
            # Handle list of pos
            elif pos.ndim == 2:
                if pos.shape[1] == 3:
                    natoms_pos = pos.shape[0]
                else:
                    raise ValueError('pos per atom must be 3-dimensional')
            else:
                raise ValueError('too many dimensions for pos')
        else:
            pos = np.zeros((1, 3), dtype='float64')
            natoms_pos = 1
        
        # Check natoms parameter values
        if natoms is not None:
            natoms = int(natoms)
            if ((natoms_atype == 1 or natoms_atype == natoms)
                and (natoms_pos == 1 or natoms_pos == natoms)):
                pass
            else:
                raise ValueError('natoms and length of atype/pos not compatible')
        else:
            if natoms_atype == natoms_pos:
                natoms = natoms_atype
            elif natoms_atype == 1:
                natoms = natoms_pos
            elif natoms_pos == 1:
                natoms = natoms_atype
            else:
                raise ValueError('lengths of atype and pos not compatible')
            
        # Initialize underlying private class attributes
        super(Atoms, self).__setattr__('_Atoms__natoms', natoms)
        super(Atoms, self).__setattr__('_Atoms__view', Atoms.PropertyDict(self))
        super(Atoms, self).__setattr__('_Atoms__dir', deepcopy(dir(self)))
        
        # Set properties
        if safecopy:
            self.view['atype'] = deepcopy(atype)
            self.view['pos'] = deepcopy(pos)
            for key, value in kwargs.items():
                self.view[key] = deepcopy(value)
        else:
            self.view['atype'] = atype
            self.view['pos'] = pos
            for key, value in kwargs.items():
                self.view[key] = value
    
    def model(self, prop_name=None, unit=None, prop_unit=None):
        """
        Generates a data model for the Atoms object.
        
        Parameters
        ----------
        prop_name : list, optional
            The Atoms properties to include.  If neither prop_name nor prop_unit
            are given, all system properties will be included.
        unit : list, optional
            Lists the units for each prop_name as stored in the table.  For a
            value of None, no conversion will be performed for that property.  
            If neither unit nor prop_units given, pos will be
            given in Angstroms and all other values will not be converted.
        prop_unit : dict, optional
            dictionary where the keys are the property keys to include, and
            the values are units to use. If neither unit nor prop_units given, 
            pos will be given in Angstroms and all other values will not be
            converted.

        Returns
        -------
        DataModelDict.DataModelDict
            A JSON/XML data model for the current Atoms object.  
        """
        
        # Set prop_unit if needed
        if prop_unit is None:
            if prop_name is None:
                prop_name = self.prop()
            if unit is None:
                unit = [None for i in range(len(prop_name))]
            
            if len(unit) != len(prop_name):
                raise ValueError('')
            prop_unit = {}
            for p, u  in zip(prop_name, unit):
                prop_unit[p] = u
        
        elif prop_name is not None or unit is not None:
            raise ValueError('prop_unit cannot be given with prop_name or unit')
        
        # Set default pos unit
        if 'pos' in prop_unit and prop_unit['pos'] is None:
            prop_unit['pos'] = 'angstrom'
        
        # Generate DataModelDict
        model = DM()
        model['atoms'] = DM()
        model['atoms']['natoms'] = self.natoms
        for prop in prop_unit:
            unit = prop_unit.get(prop, None)
            propmodel = DM()
            propmodel['name'] = prop
            propmodel['data'] = uc.model(self.prop(prop), unit)
            model['atoms'].append('property', propmodel)
        
        return model

    def __str__(self):
        """string output of Atoms data"""
        atype = self.atype # pylint: disable=no-member
        pos = self.pos # pylint: disable=no-member
        lines = ['per-atom properties = ' + str(self.prop())]
        lines.append('     id |   atype |  pos[0] |  pos[1] |  pos[2]')
        for i in range(self.natoms):
            lines.append('%7i | %7i | %7.3f | %7.3f | %7.3f' % (i, atype[i], pos[i,0], pos[i,1], pos[i,2]))
        
        return '\n'.join(lines)
    
    def __setattr__(self, name, value):
        """Control setting of user-defined attributes"""
        if not hasattr(self, name) or name in self.view:
            self.view[name] = value
        else:
            super(Atoms, self).__setattr__(name, value)
    
    def __intslice(self, intnum):
        if intnum == -1:
            return slice(intnum, None)
        else:
            return slice(intnum, intnum+1)
    
    def __deepcopy__(self, memo):
        """Properly handle deepcopy"""
        d = OrderedDict()
        atype = deepcopy(self.view['atype'])
        pos = deepcopy(self.view['pos'])
        for key in self.view:
            if key not in ['atype', 'pos']:
                d[key] = deepcopy(self.view[key])
        return Atoms(atype=atype, pos=pos, **d)
    
    def __getitem__(self, index):
        """Index getting of Atoms."""
        view = OrderedDict()
        if isinstance(index, (int, np.integer)):
            index = self.__intslice(index)
        for key in self.view.keys():
            view[key] = self.view[key][index]
        return Atoms(**view)
        
    def __setitem__(self, index, value):
        """Index setting of Atoms."""
        try:
            assert isinstance(value, Atoms)
            assert sorted(value.view.keys()) == sorted(self.view.keys())
        except:
            raise ValueError('Can only set Atoms with matching properties')
        
        if isinstance(index, (int, np.integer)):
            index = self.__intslice(index)
            
        for key in self.view.keys():
            self.view[key][index] = value.view[key]
    
    def __len__(self):
        """len of atoms = natoms."""
        return self.natoms
    
    @property
    def natoms(self):
        """int : The number of atoms in the Atoms class."""
        return self.__natoms # pylint: disable=no-member
    
    @property
    def atypes(self):
        """tuple : List of int atom types."""
        return tuple(range(1, self.natypes+1))
    
    @property
    def natypes(self):
        """int : The number of atom types in the Atoms class."""
        if np.min(self.atype) < 1: # pylint: disable=no-member
            raise ValueError('atype values < 1 not allowed')

        return int(np.max(self.atype)) # pylint: disable=no-member
        
    @property
    def view(self):
        """PropertyDict : All assigned per-atom properties."""
        return self.__view # pylint: disable=no-member
    
    def prop(self, key=None, index=None, value=None, a_id=None):
        """
        Accesses the per-atom properties for controlled getting and setting.
        For getting values, prop() always returns a copy of the underlying
        data as opposed to the data itself.
        
        Parameters
        ----------
        key : str, optional
            Per-atom property name.
        index : int, list, slice, optional
            Index of atoms.
        value : any, optional
            Property values to assign.
        a_id : int, optional
            Integer atom index.  Left in for backwards compatibility.
        
        Returns
        -------
        list
            If no parameters given, returns a list of all assigned property
            keys.
        atomman.Atoms
            If index or a_id is given without value or key, returns a new
            Atoms instance for the specified atom indices.
        numpy.ndarray
            If key (and index/a_id) is given without value, returns a copy of
            the data associated with that property key.
        
        raises
        ------
        ValueError
            If a_id and index are both given, or only value is given.
        """
        
        # Handle a_id
        if a_id is not None:
            if index is not None:
                raise ValueError('a_id and index cannot both be given')
            index = a_id
        
        # Get values if value is None
        if value is None:
            if key is None:
                # Return list of property names when no parameters are given
                if index is None:
                    return list(self.view.keys())
                # Return copy of slice if only index is given
                else:
                    return deepcopy(self[index])
            # If key is given, return copy of property values
            else:
                if index is None:
                    return deepcopy(self.view[key])
                else:
                    return deepcopy(self.view[key][index])
        
        # Set values if value is given
        else:
            # If no key, set value to atoms (or a slice)
            if key is None:
                if not isinstance(value, Atoms):
                    raise TypeError('If key is None, value must be instance of atomman.Atoms')
                if index is None:
                    self[:] = value
                else:
                    self[index] = value
            
            # If key is given, set copy of value to property
            else:
                if index is None:
                    self.view[key] = deepcopy(value)
                else:
                    self.view[key][index] = value
    
    def prop_atype(self, key, value, atype=None):
        """
        Allows for per-atom properties to be assigned according to
        Atoms.atypes.

        Parameters
        ----------
        key : str
            Per-atom property name.
        value : list, any
            Property value(s) to assign.  If atype is not given, this should be
            an object of length Atoms.natypes. Otherwise, should be a single per-atom
            value.
        atype : int, optional
            A specific atype to assign value to.

        raises
        ------
        ValueError
            If length of value does not match Atoms.natypes or atype is not in
            Atoms.atypes.
        """

        # Set values across all atype values
        if atype is None:
            value = np.asarray(value)
            if len(value) >= self.natypes:
                self.view[key] = value[self.atype - 1] # pylint: disable=no-member
            else:
                raise ValueError('length of value less than natypes')
        
        # Set values for only one atype
        else:
            if atype in self.atypes:
                if key not in self.prop():
                    self.view[key] = np.zeros_like(value)
                self.view[key][self.atype==atype] = value # pylint: disable=no-member
            else:
                raise ValueError('atype not found')
    
    def df(self):
        """
        Returns a pandas.DataFrame of all atomic properties.  Multi-dimensional
        per-atom data will be converted into multiple table columns.
        """
        
        # Initialize new dictionary of values
        values = OrderedDict()
        for key in self.view.keys():
            
            # Flatten multidimensional arrays
            for index, istr in indexstr(self.view[key].shape[1:]):
                newkey = key + istr
                
                # Copy values over
                if index == ():
                    values[newkey] = self.view[key]
                else:
                    values[newkey] = self.view[key][(Ellipsis, ) + index]
        
        # Return DataFrame
        return pd.DataFrame(values)
    
    def extend(self, value):
        """
        Allows additional atoms to be added to the end of the atoms list.
        
        Parameters
        ----------
        value : atomman.Atoms or int
            An int value will result in the atoms object being extended by
            that number of atoms, with all per-atom properties having default
            values (atype = 1, everything else = 0).  For an Atoms value, the
            current atoms list will be extended by the correct number of atoms
            and all per-atom properties in value will be copied over.  Any
            properties defined in one Atoms object and not the other will be
            set to default values.
        
        Returns
        -------
        atomman.Atoms
            A new Atoms object containing all atoms and properties of the
            current object plus the additional atoms.
        """
        
        # Handle different value types
        if isinstance(value, (int, np.integer)):
            natoms = value
            atoms = Atoms(natoms=natoms)
        elif isinstance(value, Atoms):
            natoms = value.natoms
            atoms = value        
        else:
            raise TypeError('can only add Atoms or an int # of atoms')
        
        # Generate newatoms by copying values of self + natoms extras
        index = list(range(self.natoms)) + [0 for i in range(natoms)]
        newatoms = self[index]
        
        # Create empty values for atoms.props not in newatoms (self)
        for prop in atoms.prop():
            if prop not in newatoms.prop():
                newatoms.view[prop] = np.zeros((newatoms.natoms, ) + atoms.view[prop][0].shape, dtype=atoms.view[prop][0].dtype)
        
        # Copy values to the extra atoms in newatoms
        for prop in newatoms.prop():
            if prop in atoms.prop():
                newatoms.view[prop][self.natoms:] = atoms.view[prop]
            else:
                newatoms.view[prop][self.natoms:] = np.zeros((natoms, ) + self.view[prop][0].shape, dtype=self.view[prop][0].dtype)
        
        return newatoms


# coding: utf-8

from . import Database
def load_lammps_potential(id=None, key=None, potid=None, potkey=None,
                   status='active', pair_style=None, elements=None,
                   symbols=None, verbose=False, getfiles=False,
                   database=None,
                   localpath=None, local=True, remote=True):
    """
    Gets a LAMMPS potential from the iprPy library or by downloading from
    potentials.nist.gov if a local copy is not found.  Will issue a prompt if
    multiple LAMMPS potentials match the parameters given.
    
    Parameters
    ----------
    id : str or list, optional
        The id value(s) to limit the search by.
    key : str or list, optional
        The key value(s) to limit the search by.
    potid : str or list, optional
        The potid value(s) to limit the search by.
    potkey : str or list, optional
        The potkey value(s) to limit the search by.
    status : str or list, optional
        The status value(s) to limit the search by.
    pair_style : str or list, optional
        The pair_style value(s) to limit the search by.
    elements : str or list, optional
        The included elemental model(s) to limit the search by.
    symbols : str or list, optional
        The included symbol model(s) to limit the search by.
    verbose: bool, optional
        If True, informative print statements will be used.
    getfiles : bool, optional
        If True, then the parameter files for the matching potentials
        will also be retrieved and copied to the working directory.
        If False (default) and the parameter files are in the library,
        then the returned objects' pot_dir path will be set appropriately.
    database : atomman.library.Database, optional
        A pre-existing Database class to use for accessing the records.  Useful
        if multiple potentials are being retrieved as records can be pre-loaded
        for efficiency.  If not given, a new database object will be
        initialized based on the parameters below.
    localpath : str, optional
        The path to the local database directory to use.  If not given,  will
        use the default potentials database path.  Ignored if database is
        given.
    local : bool, optional
        Indicates if the local version of the library should be searched.
        Default value is True.  Ignored if database is given.
    remote : bool, optional
        Indicates if the remote version of the library should be searched.
        Default value is True.  Ignored if database is given.
        
    Returns
    -------
    potentials.PotentialLAMMPS
        The potential object to use.

    Raises
    ------
    ValueError
        If no matching LAMMPS potentials are found.
    """
    # Create Database object and load if needed
    if database is None:
        if local is True:
            database = Database(load='lammps_potentials', status=status,
                                localpath=localpath, local=local,
                                remote=remote, verbose=verbose)
        else:
            database = Database(local=local, remote=remote, verbose=verbose)
    
    matches = database.get_lammps_potentials(id=id, key=key,
                                             potid=potid, potkey=potkey,
                                             pair_style=pair_style, status=status,
                                             elements=elements, symbols=symbols,
                                             verbose=verbose, getfiles=False)

    if len(matches) == 1:
        potential = matches[0]
    elif len(matches) > 1:
        print('Multiple matching LAMMPS potentials found')
        for i, match in enumerate(matches):
            print(i+1, match.id)
        index = int(input('Please select one:')) - 1
        potential = matches[index]
    else:
        raise ValueError('No matching LAMMPS potentials found')

    if getfiles is True:
        return database.get_lammps_potential(id=potential.id, getfiles=True)
    else:
        return potential


# coding: utf-8
# Standard Python imports
from collections import OrderedDict
import os

# http://pandas.pydata.org/
import pandas as pd

# http://www.numpy.org/
import numpy as np

# https://www.scipy.org/
from scipy.interpolate import griddata, Rbf, NearestNDInterpolator

# http://matplotlib.org/
import matplotlib.pyplot as plt

# https://github.com/usnistgov/DataModelDict
from DataModelDict import DataModelDict as DM

# atomman imports
from .. import Box
from .. import unitconvert as uc
from ..tools import miller
from ..mep import create_path

class GammaSurface(object):
    """
    Class for representing gamma surfaces, i.e., generalized stacking faults.
    """
    
    def __init__(self, model=None, a1vect=None, a2vect=None,
                 a1=None, a2=None, E_gsf=None, box=None, delta=None):
        """
        Class initializer. Parameter model must be given alone. Otherwise,
        all or none of a1vect, a2vect, a1, a2, and E_gsf must be given.
        
        Parameters
        ----------
        model : str, file-like object, DataModelDict, optional
            XML/JSON data model containing the stacking fault information.
        a1vect : array-like object, optional
            The a1 shifting vector.  If box is given, a1vect is taken as a
            crystal lattice vector, otherwise as a Cartesian vector.
        a2vect : array-like object, optional
            The a2 shifting vector.  If box is given, a2vect is taken as a
            crystal lattice vector, otherwise as a Cartesian vector.
        a1 : array-like object, optional
            List of fractional coordinates along a1vect corresponding to the
            E_gsf (and delta) values.
        a2 : array-like object, optional
            List of fractional coordinates along a2vect corresponding to the
            E_gsf (and delta) values.
        E_gsf : array-like object, optional
            List of generalized stacking fault energies for the positions
            associated with the corresponding (a1, a2) fractional coordinates.
        box : atomman.Box, optional
            Defines unit cell box dimensions for conversion between crystal
            lattice and Cartesian vectors.  If not given, will be set as a
            square unit box, thus no conversion will occur (i.e. a1vect,
            a2vect will be Cartesian).
        delta : array-like object, optional
            List of change in displacements normal to the fault plane for the
            positions associated with the corresponding (a1, a2) fractional
            coordinates.
        """
    
        # Load model if given
        if model is not None:
            try:
                assert box is None
                assert a1vect is None
                assert a2vect is None
                assert a1 is None
                assert a2 is None
                assert E_gsf is None
                assert delta is None
            except:
                raise TypeError('model cannot be given with any other parameter')
            else:
                self.model(model=model)
        
        # Set values if given
        elif (a1 is not None or a2 is not None or E_gsf is not None
              or a1vect is not None or a2vect is not None
              or delta is not None or box is not None):
            try:
                assert a1vect is not None
                assert a2vect is not None
                assert a1 is not None
                assert a2 is not None
                assert E_gsf is not None
            except:
                raise TypeError('Defining data requires a1vect, a2vect, a1, a2 and E_gsf')
            else:  
                self.set(a1vect, a2vect, a1, a2, E_gsf,  box=box, delta=delta)
        
        # Set flag for no supplied data
        else:
            self.__hasdata = False
    
    @property
    def data(self):
        """pandas.DataFrame : The raw data."""
        if self.__hasdata:
            return self.__data
        else:
            raise AttributeError('gamma surface data not set')
          
    @property
    def a1vect(self):
        """numpy.ndarray : The a1 shifting vector."""
        if self.__hasdata:
            return self.__a1vect
        else:
            raise AttributeError('gamma surface data not set')
    
    @property
    def a2vect(self):
        """numpy.ndarray : The a2 shifting vector."""
        if self.__hasdata:
            return self.__a2vect
        else:
            raise AttributeError('gamma surface data not set')
    
    @property
    def planenormal(self):
        """numpy.ndarray : The Cartesian vector normal to the fault plane."""
        if self.__hasdata:
            return self.__planenormal
        else:
            raise AttributeError('gamma surface data not set')

    @property
    def box(self):
        """
        atomman.Box : A unit cell box used for converting between
                      crystal lattice and Cartesian vectors.
        """
        if self.__hasdata:
            return self.__box
        else:
            raise AttributeError('gamma surface data not set')
    
    def set(self, a1vect, a2vect, a1, a2, E_gsf,  box=None, delta=None):
        """
        Sets generalized stacking fault data.
        
        Parameters
        ----------
        a1vect : array-like object
            The a1 shifting vector.  If box is given, a1vect is taken as a
            crystal lattice vector, otherwise as a Cartesian vector.
        a2vect : array-like object
            The a2 shifting vector.  If box is given, a2vect is taken as a
            crystal lattice vector, otherwise as a Cartesian vector.
        a1 : array-like object
            List of fractional coordinates along a1vect corresponding to the
            E_gsf (and delta) values.
        a2 : array-like object
            List of fractional coordinates along a2vect corresponding to the
            E_gsf (and delta) values.
        E_gsf : array-like object
            List of generalized stacking fault energies for the positions
            associated with the corresponding (a1, a2) fractional coordinates.
        box : atomman.Box, optional
            Defines unit cell box dimensions for conversion between crystal
            lattice and Cartesian vectors.  If not given, will be set as a
            square unit box, thus no conversion will occur (i.e. a1vect,
            a2vect will be Cartesian).
        delta : array-like object, optional
            List of change in displacements normal to the fault plane for the
            positions associated with the corresponding (a1, a2) fractional
            coordinates.
        """
        # Set a1vect
        if isinstance(a1vect, str):
            a1vect = a1vect.split()
        a1vect = np.asarray(a1vect, dtype=float)
        if a1vect.shape == (4,):
            a1vect = miller.vector4to3(a1vect)
        elif a1vect.shape != (3,):
            raise ValueError('a1vect must be a 3D vector')
        self.__a1vect = a1vect

        # Set a2vect
        if isinstance(a2vect, str):
            a2vect = a2vect.split()
        a2vect = np.asarray(a2vect, dtype=float)
        if a2vect.shape == (4,):
            a2vect = miller.vector4to3(a2vect)
        elif a2vect.shape != (3,):
            raise ValueError('a2vect must be a 3D vector')
        self.__a2vect = a2vect

        # Set box
        if box is None:
            box = Box()
        if not isinstance(box, Box):
            raise TypeError('box must be an atomman.Box')
        self.__box = box

        # Set plane normal
        a1vect = np.dot(a1vect, box.vects)
        a2vect = np.dot(a2vect, box.vects)
        planenormal = np.cross(a1vect, a2vect)
        self.__planenormal = planenormal / np.linalg.norm(planenormal)

        # Set data
        data = OrderedDict()
        data['a1'] = a1
        data['a2'] = a2
        data['E_gsf'] = E_gsf
        if delta is not None:
            data['delta'] = delta
        self.__data = pd.DataFrame(data)

        # Fit
        self.__hasdata = True
        self.fit()        
    
    def fit(self):
        """
        Defines the interpolation functions from the raw data.
        """
        
        # Ignore a1, a2=1.0 values if included
        shortdata = self.data[~(np.isclose(self.data.a1, 1.0) | np.isclose(self.data.a2, 1.0))]
        
        # Create supercells of values
        a1 = shortdata.a1
        a2 = shortdata.a2
        a1 = np.concatenate([a1-1, a1-1, a1-1, a1, a1, a1, a1+1, a1+1, a1+1])
        a2 = np.concatenate([a2-1, a2, a2+1, a2-1, a2, a2+1, a2-1, a2, a2+1])

        # Find values in 0-1 cell +- one point
        ua1 = np.unique(a1)
        ua2 = np.unique(a2)
        a1min = ua1[np.where(np.isclose(ua1, 0.0))[0][0] - 1] - 1e-8
        a1max = ua1[np.where(np.isclose(ua1, 1.0))[0][-1] + 1] + 1e-8
        a2min = ua2[np.where(np.isclose(ua2, 0.0))[0][0] - 1] - 1e-8
        a2max = ua2[np.where(np.isclose(ua2, 1.0))[0][-1] + 1] + 1e-8
        ix = np.where((a1 >= a1min) & (a1 <= a1max) & (a2 >= a2min) & (a2 <= a2max))
        
        # Fit energy
        E_gsf = np.concatenate([shortdata.E_gsf] * 9)
        self.__E_gsf_fit = Rbf(a1[ix], a2[ix], E_gsf[ix])
        self.__E_gsf_nearest = NearestNDInterpolator(np.array([a1[ix], a2[ix]]).T, E_gsf[ix])
        
        # Fit delta
        if 'delta' in self.data:
            delta = np.concatenate([shortdata.delta] * 9)
            self.__delta_fit = Rbf(a1[ix], a2[ix], delta[ix])
            self.__delta_nearest = NearestNDInterpolator(np.array([a1[ix], a2[ix]]).T, delta[ix])
    
    def model(self, model=None, length_unit='angstrom',
              energyperarea_unit='mJ/m^2'):
        """
        Return or set DataModelDict representation of the gamma surface.
        
        Parameters
        ----------
        model : str, file-like object or DataModelDict, optional
            XML/JSON content to extract gamma surface energy from. If not
            given, model content will be generated.
        length_unit : str, optional
            Units to report delta displacement values in when a new model is
            generated. Default value is 'angstrom'.
        energyperarea_unit : str, optional
            Units to report fault energy values in when a new model is
            generated.  Default value is 'mJ/m^2'.
        
        Returns
        -------
        DataModelDict
            A dictionary containing the stacking fault data of the
            GammaSurface object.  Returned if model is not given.
        """
        # Set values if model given
        if model is not None:
            model = DM(model).find('stacking-fault-map')
            
            # Read in box, a1vect and a2vect
            box = Box(avect = model['box']['avect'],
                      bvect = model['box']['bvect'],
                      cvect = model['box']['cvect'])
            
            a1vect = model['shift-vector-1']
            a2vect = model['shift-vector-2']
            
            # Read in stacking fault data
            gsf = model.find('stacking-fault-relation')
            
            a1 = gsf['shift-vector-1-fraction']
            a2 = gsf['shift-vector-2-fraction']
            E_gsf = uc.value_unit(gsf['energy'])
            try:
                delta = uc.value_unit(gsf['plane-separation'])
            except:
                delta = None
            self.set(a1vect, a2vect, a1, a2, E_gsf, box=box, delta=delta)
        
        # Generate model
        else:
            model = DM()
            model['stacking-fault-map'] = sfm = DM()
            sfm['box'] = DM()
            sfm['box']['avect'] = list(self.box.avect)
            sfm['box']['bvect'] = list(self.box.bvect)
            sfm['box']['cvect'] = list(self.box.cvect)
            sfm['shift-vector-1'] = list(self.a1vect)
            sfm['shift-vector-2'] = list(self.a2vect)
            sfm['stacking-fault-relation'] = sfr = DM()
            sfr['shift-vector-1-fraction'] = list(self.data.a1)
            sfr['shift-vector-2-fraction'] = list(self.data.a2)
            sfr['energy'] = uc.model(self.data.E_gsf, energyperarea_unit)
            if 'delta' in self.data:
                sfr['plane-separation'] = uc.model(self.data.delta, length_unit)
            
            return model
    
    def a12_to_pos(self, a1, a2, a1vect=None, a2vect=None):
        """
        Conversion function from normalized a1, a2 coordinates to Cartesian
        positions.
        
        Parameters
        ----------
        a1 : float(s)
            Fractional distance(s) along a1 vector.
        a2 : float(s)
            Fractional distance(s) along a2 vector.
        a1vect : np.array, optional
            Crystal vector for the a1 vector.  Default value of None uses the
            saved a1vect.
        a2vect : np.array, optional
            Crystal vector for the a2 vector.  Default value of None uses the
            saved a2vect.
        
        Returns
        -------
        np.array
            3D Cartesian position vector(s).
        """
        # Handle a1vect and a2vect
        if a1vect is None:
            a1vect = self.a1vect
        a1vect = np.asarray(a1vect)

        if a2vect is None:
            a2vect = self.a2vect
        a2vect = np.asarray(a2vect)
        
        # Convert a1vect and a2vect from crystal to Cartesian coordinates
        a1vect = np.dot(a1vect, self.box.vects)
        a2vect = np.dot(a2vect, self.box.vects)
        
        # Transform a1, a2 to Cartesian pos
        return np.outer(a1, a1vect) + np.outer(a2, a2vect)
        
    def pos_to_xy(self, pos, xvect=None):
        """
        Conversion function from Cartesian positions to plotting x, y
        coordinates.
        
        Parameters
        ----------
        pos: np.array
            3D Cartesian position vector(s).
        xvect : np.array, optional
            Cartesian vector corresponding to the plotting x-axis. If None (default), this is
            taken as the Cartesian of a1vect.
        
        Returns
        -------
        x : float(s)
            Plotting x coordinate(s).
        y : float(s)
            Plotting y coordinate(s).
        """
        # Handle xvect
        if xvect is None:
            xvect = np.dot(self.a1vect, self.box.vects)
        xvect = np.asarray(xvect)        
        if not np.isclose(np.dot(xvect, self.planenormal), 0.0):
            raise ValueError('xvect must be in plane defined by a1vect and a2vect')
        
        # Build transformation tensor
        yvect = np.cross(self.planenormal, xvect)
        transform = np.array([xvect, yvect, self.planenormal])
        transform = (transform.T / np.linalg.norm(transform, axis=1)).T
        
        # Transform coordinates to x,y,z orientation
        pos = transform.dot(pos.T).T
        
        # Return x, y coordinates
        return pos[...,0], pos[...,1]
    
    def a12_to_xy(self, a1, a2, a1vect=None, a2vect=None, xvect=None):
        """
        Conversion function from normalized a1, a2 coordinates to plotting x, y
        coordinates.
        
        Parameters
        ----------
        a1 : float(s)
            Fractional distance(s) along a1 vector.
        a2 : float(s)
            Fractional distance(s) along a2 vector.
        a1vect : np.array, optional
            Crystal vector for the a1 vector.  Default value of None uses the
            saved a1vect.
        a2vect : np.array, optional
            Crystal vector for the a2 vector.  Default value of None uses the
            saved a2vect.
        xvect : np.array, optional
            Cartesian vector corresponding to the plotting x-axis. If None (default), this is
            taken as the Cartesian of a1vect.
        
        Returns
        -------
        x : float(s)
            Plotting x coordinate(s).
        y : float(s)
            Plotting y coordinate(s).
        """
        # Set xvect as given a1vect if needed
        if a1vect is not None and xvect is None:
            xvect = np.dot(a1vect, self.box.vects)
        
        # Transform from a1, a2 to pos
        pos = self.a12_to_pos(a1, a2, a1vect=a1vect, a2vect=a2vect)

        # Transform from pos to x, y
        return self.pos_to_xy(pos, xvect=xvect)
    
    def pos_to_a12(self, pos, a1vect=None, a2vect=None):
        """
        Conversion function from Cartesian positions to normalized a1, a2
        coordinates.
        
        Parameters
        ----------
        pos : np.array
            3D Cartesian position vector(s).
        a1vect : np.array, optional
            Crystal vector for the a1 vector.  Default value of None uses the
            saved a1vect.
        a2vect : np.array, optional
            Crystal vector for the a2 vector.  Default value of None uses the
            saved a2vect.

        Returns
        -------
        a1 : float(s)
            Fractional distance(s) along a1 vector.
        a2 : float(s)
            Fractional distance(s) along a2 vector.
        """

        # Handle a1vect and a2vect
        if a1vect is None:
            a1vect = self.a1vect
        a1vect = np.asarray(a1vect)

        if a2vect is None:
            a2vect = self.a2vect
        a2vect = np.asarray(a2vect)
        
        # Convert a1vect and a2vect from crystal to Cartesian coordinates
        a1vect = np.dot(a1vect, self.box.vects)
        a2vect = np.dot(a2vect, self.box.vects)
        
        # Solve for a1, a2, a3
        a3vect = np.cross(a1vect, a2vect)
        coeffs = np.array([a1vect, a2vect, a3vect]).T
        if pos.ndim == 2:
            coeffs = np.array([coeffs])
        a123 = np.linalg.solve(coeffs, pos)
        assert np.allclose(a123[...,2], 0.0, atol=1e-6), np.abs(a123[...,2]).max()

        # Return a1, a2
        return a123[...,0], a123[...,1]
    
    def xy_to_pos(self, x, y, xvect=None):
        """
        Conversion function from plotting x, y coordinates to Cartesian
        positions.
        
        Parameters
        ----------
        x : float(s)
            Plotting x coordinate(s).
        y : float(s)
            Plotting y coordinate(s).
        xvect : np.array, optional
            Cartesian vector corresponding to the plotting x-axis. If None
            (default), this is taken as the Cartesian of a1vect.
        
        Returns
        -------
        pos: np.array
            3D Cartesian position vector(s).
        """
        # Assign default xvect if needed
        if xvect is None:
            xvect = np.dot(self.a1vect, self.box.vects)
        xvect = np.asarray(xvect)
        if not np.isclose(np.dot(xvect, self.planenormal), 0.0):
            raise ValueError('xvect must be in plane defined by a1vect and a2vect')
       
        # Build transformation tensor
        yvect = np.cross(self.planenormal, xvect)
        transform = np.array([xvect, yvect, self.planenormal])
        transform = (transform.T / np.linalg.norm(transform, axis=1)).T
        transform = np.linalg.inv(transform)
        
        # Transform coords
        pos = np.outer(x, [1,0,0]) + np.outer(y, [0,1,0])
        pos = transform.dot(pos.T).T
        
        # Return Cartesian pos
        return pos
    
    def xy_to_a12(self, x, y, a1vect=None, a2vect=None, xvect=None):
        """
        Conversion function from plotting x, y coordinates to normalized a1, a2
        coordinates.
        
        Parameters
        ----------
        x : float(s)
            Plotting x coordinate(s).
        y : float(s)
            Plotting y coordinate(s).
        a1vect : np.array, optional
            Crystal vector for the a1 vector.  Default value of None uses the
            saved a1vect.
        a2vect : np.array, optional
            Crystal vector for the a2 vector.  Default value of None uses the
            saved a2vect.
        xvect : np.array, optional
            Cartesian vector corresponding to the plotting x-axis. If None
            (default), this is taken as the Cartesian of a1vect.
        
        Returns
        -------
        a1 : float(s)
            Fractional distance(s) along a1 vector.
        a2 : float(s)
            Fractional distance(s) along a2 vector.
        """
        
        # Set xvect to given a1vect if needed
        if a1vect is not None and xvect is None:
            xvect = np.dot(a1vect, self.box.vects)

        # Convert x, y to pos    
        pos = self.xy_to_pos(x, y, xvect=xvect)

        # Convert pos to a1, a2
        return self.pos_to_a12(pos, a1vect=a1vect, a2vect=a2vect)
    
    def E_gsf(self, **kwargs):
        """
        Returns values for generalized stacking fault energy interpolated from
        the raw data.  Values can be obtained relative to a1, a2 fractional
        coordinates, x, y plotting coordinates, or pos Cartesian coordinates.
        
        Parameters
        ----------
        a1 : float(s), optional
            Fractional coordinate(s) along a1vect.
        a2 : float(s), optional
            Fractional coordinate(s) along a2vect.
        pos : np.array, optional
            3D Cartesian position vector(s).
        x : float(s), optional
            Plotting x coordinate(s).
        y : float(s), optional
            Plotting y coordinate(s).
        a1vect : np.array, optional
            Vector for the a1 fractional coordinates.  Default value of None 
            uses the saved a1vect.
        a2vect : np.array, optional
            Vector for the a2 fractional coordinates.  Default value of None 
            uses the saved a2vect.
        xvect : np.array, optional
            Cartesian vector corresponding to the plotting x-axis. If None
            (default), this is taken as the Cartesian of a1vect.
        smooth : bool, optional
            If True (default) the returned values are smoothed using a RBF fit.
            If False, the closest measured values are returned.
        """
        if not self.__hasdata:
            raise AttributeError('gamma surface data not set')
        
        smooth = kwargs.pop('smooth', True)

        # Convert x, y to a1, a2
        if 'x' in kwargs:
            x = kwargs.pop('x')
            y = kwargs.pop('y')
            a1vect = kwargs.pop('a1vect', None)
            a2vect = kwargs.pop('a2vect', None)
            xvect = kwargs.pop('xvect', None)
            assert len(kwargs) == 0, 'Unknown/incompatible arguments given'
            a1, a2 = self.xy_to_a12(x, y, a1vect=a1vect, a2vect=a2vect, xvect=xvect)
        
        # Convert pos to a1, a2
        elif 'pos' in kwargs:
            pos = kwargs.pop('pos')
            a1vect = kwargs.pop('a1vect', None)
            a2vect = kwargs.pop('a2vect', None)
            assert len(kwargs) == 0, 'Unknown/incompatible arguments given'
            a1, a2 = self.pos_to_a12(pos, a1vect=a1vect, a2vect=a2vect)
        
        # Get a1, a2 from kwargs
        else:
            a1 = np.array(kwargs.pop('a1'))
            a2 = np.array(kwargs.pop('a2'))
            a1vect = kwargs.pop('a1vect', None)
            a2vect = kwargs.pop('a2vect', None)
            assert len(kwargs) == 0, 'Unknown/incompatible arguments given'
            if a1vect is not None or a2vect is not None:
                shape = a1.shape
                # Convert into pos using given a1vect, a2vect, then back into a1, a2
                pos = self.a12_to_pos(a1, a2, a1vect=a1vect, a2vect=a2vect)
                a1, a2 = self.pos_to_a12(pos)
                a1 = a1.reshape(shape)
                a2 = a2.reshape(shape)

        # Return interpolated values
        if smooth:
            cushion = (1 - self.data.a1.max()) / 2
    
            # Wrap all a1, a2 values within [-cushion, 1.0 - cushion)
            while np.any(a1 >= 1.0 - cushion): 
                a1[a1 >= 1.0 - cushion] -= 1.0
            while np.any(a1 < -cushion): 
                a1[a1 < -cushion] += 1.0
            while np.any(a2 >= 1.0 - cushion): 
                a2[a2 >= 1.0 - cushion] -= 1.0
            while np.any(a2 < -cushion): 
                a2[a2 < -cushion] += 1.0
            
            # Compute weighting factors
            def zone1(x):
                return (x + cushion) / (2 * cushion)
            def zone2(x):
                return np.ones_like(x)
            x = np.piecewise(a1, [a1 < cushion, a1>= cushion], [zone1, zone2])
            y = np.piecewise(a2, [a2 < cushion, a2>= cushion], [zone1, zone2])
            
            # Linear smoothing across boundaries
            return ( x * y * self.__E_gsf_fit(a1, a2)
                + x * (1 - y) * self.__E_gsf_fit(a1, a2 + 1)
                + (1 - x) * y * self.__E_gsf_fit(a1 + 1, a2)
                + (1 - x) * (1 - y) * self.__E_gsf_fit(a1 + 1, a2 + 1))
        
        # Return nearest values
        else:

            # Wrap all values within 0.0 < a1, a2 < 1.0
            while np.any(a1 > 1.0): 
                a1[a1 > 1.0] -= 1.0
            while np.any(a1 < 0.0): 
                a1[a1 < 0.0] += 1.0
            while np.any(a2 > 1.0): 
                a2[a2 > 1.0] -= 1.0
            while np.any(a2 < 0.0): 
                a2[a2 < 0.0] += 1.0

            return self.__E_gsf_nearest(np.array([a1.flatten(), a2.flatten()]).T).reshape(a1.shape)
    
    def delta(self, **kwargs):
        """
        Returns values for generalized stacking fault energy interpolated from
        the raw data.  Values can be obtained relative to a1, a2 fractional
        coordinates, x, y plotting coordinates, or pos Cartesian coordinates.
        
        Parameters
        ----------
        a1 : float(s), optional
            Fractional coordinate(s) along a1vect.
        a2 : float(s), optional
            Fractional coordinate(s) along a2vect.
        pos : np.array, optional
            3D Cartesian position vector(s).
        x : float(s), optional
            Plotting x coordinate(s).
        y : float(s), optional
            Plotting y coordinate(s).
        a1vect : np.array, optional
            Vector for the a1 fractional coordinates.  Default value of None 
            uses the saved a1vect.
        a2vect : np.array, optional
            Vector for the a2 fractional coordinates.  Default value of None 
            uses the saved a2vect.
        xvect : np.array, optional
            Cartesian vector corresponding to the plotting x-axis. If None
            (default), this is taken as the Cartesian of a1vect.
        smooth : bool, optional
            If True (default) the returned values are smoothed using a RBF fit.
            If False, the closest measured values are returned.
        """
        if not self.__hasdata:
            raise AttributeError('gamma surface data not set')
        if 'delta' not in self.data:
            raise AttributeError('delta data not set')
        
        smooth = kwargs.pop('smooth', True)

        # Convert x, y to a1, a2
        if 'x' in kwargs:
            x = kwargs.pop('x')
            y = kwargs.pop('y')
            a1vect = kwargs.pop('a1vect', None)
            a2vect = kwargs.pop('a2vect', None)
            xvect = kwargs.pop('xvect', None)
            assert len(kwargs) == 0, 'Unknown/incompatible arguments given'
            a1, a2 = self.xy_to_a12(x, y, a1vect=a1vect, a2vect=a2vect, xvect=xvect)
        
        # Convert pos to a1, a2
        elif 'pos' in kwargs:
            pos = kwargs.pop('pos')
            a1vect = kwargs.pop('a1vect', None)
            a2vect = kwargs.pop('a2vect', None)
            assert len(kwargs) == 0, 'Unknown/incompatible arguments given'
            a1, a2 = self.pos_to_a12(pos, a1vect=a1vect, a2vect=a2vect)
        
        # Get a1, a2 from kwargs
        else:
            a1 = np.array(kwargs.pop('a1'))
            a2 = np.array(kwargs.pop('a2'))
            a1vect = kwargs.pop('a1vect', None)
            a2vect = kwargs.pop('a2vect', None)
            assert len(kwargs) == 0, 'Unknown/incompatible arguments given'
            if a1vect is not None or a2vect is not None:
                shape = a1.shape               
                # Convert into pos using given a1vect, a2vect, then back into a1, a2
                pos = self.a12_to_pos(a1, a2, a1vect=a1vect, a2vect=a2vect)
                a1, a2 = self.pos_to_a12(pos)
                a1 = a1.reshape(shape)
                a2 = a2.reshape(shape)
        
        # Wrap all values within 0.0 < a1, a2 < 1.0
        while np.any(a1 > 1.0): 
            a1[a1 > 1.0] -= 1.0
        while np.any(a1 < 0.0): 
            a1[a1 < 0.0] += 1.0
        while np.any(a2 > 1.0): 
            a2[a2 > 1.0] -= 1.0
        while np.any(a2 < 0.0): 
            a2[a2 < 0.0] += 1.0
        
        if smooth:
            return self.__delta_fit(a1, a2)
        else:
            return self.__delta_nearest(np.array([a1.flatten(), a2.flatten()]).T).reshape(a1.shape)
    
    def E_gsf_surface_plot(self, normalize=False, smooth=True, 
                           a1vect=None, a2vect=None, xvect=None,
                           length_unit='Å', energyperarea_unit='eV/Å^2',
                           numx=100, numy=100, figsize=None, **kwargs):
        """
        Creates a 2D surface plot from the stacking fault energy values.
        
        Parameters
        ----------
        normalize : bool, optional
            Flag indicating if axes are Cartesian (False, default) or
            normalized by a1, a2 vectors (True).
        smooth : bool, optional
            If True (default), then plot shows smooth interpolated values.
            If False, plot shows nearest raw data values.
        a1vect : np.array, optional
            Crystal vector for the a1 vector to use for plotting.  Default
            value of None uses the saved a1vect.
        a2vect : np.array, optional
            Crystal vector for the a2 vector to use for plotting.  Default
            value of None uses the saved a2vect.
        xvect : numpy.array, optional
            Crystal vector to align with the plotting x-axis for 
            non-normalized plots.  If not given, this is taken as the Cartesian
            of a1vect.
        length_unit : str, optional
            The unit of length to display non-normalized axes values in.
            Default value is 'Å'.
        energyperarea_unit : str, optional
            The unit of energy per area to display the stacking fault energies
            in. Default value is 'eV/Å^2'.
        numx : int, optional
            The number of plotting points to use along the x-axis.  Default
            value is 100.
        numy : int, optional
            The number of plotting points to use along the y-axis.  Default
            value is 100.       
        figsize : tuple or None, optional
            The figure's x,y dimensions.  If None (default), the values are
            scaled such that the x,y spacings are approximately equal, and the
            larger of the two values is set to 10.
        **kwargs : dict, optional
            Additional keywords are passed into the underlying 
            matplotlib.pyplot.pcolormesh(). This allows control of such things
            like the colormap (cmap).
            
        Returns
        -------
        matplotlib.figure
        """
        if not self.__hasdata:
            raise AttributeError('gamma surface data not set')

        # Extract data
        if a1vect is None:
            a1vect = self.a1vect
        a1vect = np.asarray(a1vect)
        if a2vect is None:
            a2vect = self.a2vect
        a2vect = np.asarray(a2vect)

        # Generate grids of a1, a2 values from numx, numy
        x_grid, y_grid = np.meshgrid(np.linspace(0, 1, numx),
                                     np.linspace(0, 1, numy))
        
        # Generate grid of values either with or without interpolation
        C = self.E_gsf(a1=x_grid, a2=y_grid, a1vect=a1vect, a2vect=a2vect, smooth=smooth)
        
        # Convert units of C using energyperarea_unit
        C = uc.get_in_units(C, energyperarea_unit)
        
        # Set parameters for normalized plots
        if normalize is True:
            yscale = 1
            xlabel = f'$a_1$ = {a1vect}'
            ylabel = f'$a_2$ = {a2vect}'
        
        # Set parameters for absolute plots
        else:
            shape = x_grid.shape
            x_grid, y_grid = self.a12_to_xy(x_grid.flatten(), y_grid.flatten(),
                                            a1vect=a1vect, a2vect=a2vect, xvect=xvect)
            x_grid.shape = shape
            y_grid.shape = shape
            x_grid = uc.get_in_units(x_grid, length_unit)
            y_grid = uc.get_in_units(y_grid, length_unit)
            yscale = (y_grid.max()-y_grid.min()) / (x_grid.max() - x_grid.min())
            xlabel = f'$x$ along {a1vect} (${length_unit}$)'
            ylabel = f'$y$ along {a2vect} (${length_unit}$)'
        
        # Set default figsize if needed
        if figsize is None:
            xscale = 1.175
            if yscale < 1:
                figsize = (xscale * 10, 10 * yscale)
            else:
                figsize = (xscale * 10 / yscale, 10)
        
        # Generate plot
        fig = plt.figure(figsize=figsize)
        plt.pcolormesh(x_grid, y_grid, C, **kwargs)
        plt.xlabel(xlabel, fontsize='xx-large')
        plt.ylabel(ylabel, fontsize='xx-large')
        cbar = plt.colorbar(aspect=40, fraction=0.1)
        cbar.ax.set_ylabel(f'$γ_{{gsf}}$ (${energyperarea_unit}$)', fontsize='x-large')
        
        return fig
    
    def E_gsf_line_plot(self, vect=None, num=None, smooth=True,
                        length_unit='Å', energyperarea_unit='eV/Å^2',
                        figsize=None, fig=None, **kwargs):
        """
        Generates a line plot for the interpolated generalized stacking fault
        energy along a specified crystallographic vector in the (a1, a2) plane.
        
        Parameters
        ----------
        vect : numpy.array, optional
            Vector to plot the gsf along.  If box is set, this vect will be a
            lattice vector, otherwise it will be a Cartesian vector.  Must be 
            in the plane defined by the GammaSurface object's a1vect and 
            a2vect vectors.  Default value will use the set a1vect.
        num : int, optional
            The number of points to evaluate the generalized stacking fault
            energy for.  Default value is 100 if smooth is True, otherwise is
            number of unique a1 values from 0 to 1.
        smooth : bool, optional
            If True (default), then plot shows smooth interpolated values.
            If False, plot shows nearest raw data values.
        length_unit : str, optional
            The unit of length to display the x-axis coordinates in.
            Default value is 'Å'.
        energyperarea_unit : str, optional
            The unit of energy per area to display the stacking fault energies
            in. Default value is 'eV/Å^2'.
        figsize : tuple, optional
            The x,y size of the figure to return.  Default value is (10, 6).
        fig : matplotlib.figure, optional
            An existing figure object to add the new plot to.  If not given, a
            new figure is generated.
        **kwargs : dict, optional
            Additional keywords are passed into the underlying 
            matplotlib.pyplot.plot(). This allows control of such things
            like line color, style, etc.
        
        Returns
        -------
        matplotlib.figure
        """
        if not self.__hasdata:
            raise AttributeError('gamma surface data not set')

        if num is None:
            if smooth:
                num = 100
            else:
                unique_a1 = np.unique([self.data.a1 - 1, self.data.a1, self.data.a1 + 1])
                num = len(unique_a1[(unique_a1 >=-0.000001) & (unique_a1 <=1.000001)])
        
        # Generate coordinates
        a1 = np.linspace(0, 1, num)
        a2 = np.zeros(num)
        pos = self.a12_to_pos(a1, a2, a1vect=vect)
        
        # Evaluate interpolated energy and distance along x
        E = uc.get_in_units(self.E_gsf(pos=pos, smooth=smooth), energyperarea_unit)
        x = uc.get_in_units(np.linalg.norm(pos, axis=1), length_unit)
        
        # Create plot
        xmax = x.max()
        emin = E.min() * 1.05
        emax = E.max() * 1.05
        if fig is None:
            if figsize is None:
                figsize = (10, 6)
            fig = plt.figure(figsize=figsize)
        else:
            old_xmax = fig.axes[0].get_xlim()[-1]
            old_emin = fig.axes[0].get_ylim()[0]
            old_emax = fig.axes[0].get_ylim()[-1]
            if old_xmax > xmax:
                xmax = old_xmax
            if old_emin < emin:
                emin = old_emin
            if old_emax > emax:
                emax = old_emax
        if 'fmt' in kwargs:
            fmt = kwargs.pop('fmt')      
            plt.plot(x, E, fmt, **kwargs)
        else:
            plt.plot(x, E, **kwargs)
        
        if vect is None:
            vect = self.a1vect

        plt.xlabel(f'$x$ along {vect} (${length_unit}$)', fontsize='x-large')
        plt.ylabel(f'$γ_{{gsf}}$ (${energyperarea_unit}$)', fontsize='x-large')
        plt.xlim(0, xmax)
        plt.ylim(emin, emax)
        
        return fig
    
    def delta_surface_plot(self, normalize=False, smooth=True, 
                           a1vect=None, a2vect=None, xvect=None,
                           length_unit='Å',
                           numx=100, numy=100, figsize=None, **kwargs):
        """
        Creates a 2D surface plot from the delta planar displacement values.
        
        Parameters
        ----------
        normalize : bool, optional
            Flag indicating if axes are Cartesian (False, default) or
            normalized by a1, a2 vectors (True).
        smooth : bool, optional
            If True (default), then plot shows smooth interpolated values.
            If False, plot shows nearest raw data values.
        a1vect : np.array, optional
            Crystal vector for the a1 vector to use for plotting.  Default
            value of None uses the saved a1vect.
        a2vect : np.array, optional
            Crystal vector for the a2 vector to use for plotting.  Default
            value of None uses the saved a2vect.
        xvect : numpy.array, optional
            Crystal vector to align with the plotting x-axis for 
            non-normalized plots.  If not given, this is taken as the Cartesian
            of a1vect.
        length_unit : str, optional
            The unit of length to display delta and non-normalized axes values
            in.  Default value is 'Å'.
        numx : int, optional
            The number of plotting points to use along the x-axis.  Default
            value is 100.
        numy : int, optional
            The number of plotting points to use along the y-axis.  Default
            value is 100.       
        figsize : tuple or None, optional
            The figure's x,y dimensions.  If None (default), the values are
            scaled such that the x,y spacings are approximately equal, and the
            larger of the two values is set to 10.
        **kwargs : dict, optional
            Additional keywords are passed into the underlying 
            matplotlib.pyplot.pcolormesh(). This allows control of such things
            like the colormap (cmap).
            
        Returns
        -------
        matplotlib.figure
        """
        if not self.__hasdata:
            raise AttributeError('gamma surface data not set')
        if 'delta' not in self.data:
            raise AttributeError('delta data not set')

        # Extract data
        if a1vect is None:
            a1vect = self.a1vect
        a1vect = np.asarray(a1vect)
        if a2vect is None:
            a2vect = self.a2vect
        a2vect = np.asarray(a2vect)

        # Generate grids of a1, a2 values from numx, numy
        x_grid, y_grid = np.meshgrid(np.linspace(0, 1, numx),
                                     np.linspace(0, 1, numy))
        
        # Generate grid of values either with or without interpolation
        C = self.delta(a1=x_grid, a2=y_grid, a1vect=a1vect, a2vect=a2vect, smooth=smooth)
        
        # Convert units of C using length_unit
        C = uc.get_in_units(C, length_unit)
        
        # Set parameters for normalized plots
        if normalize is True:
            yscale = 1
            xlabel = f'$a_1$ = {a1vect}'
            ylabel = f'$a_2$ = {a2vect}'
        
        # Set parameters for absolute plots
        else:
            shape = x_grid.shape
            x_grid, y_grid = self.a12_to_xy(x_grid.flatten(), y_grid.flatten(),
                                            a1vect=a1vect, a2vect=a2vect, xvect=xvect)
            x_grid.shape = shape
            y_grid.shape = shape
            x_grid = uc.get_in_units(x_grid, length_unit)
            y_grid = uc.get_in_units(y_grid, length_unit)
            yscale = (y_grid.max()-y_grid.min()) / (x_grid.max() - x_grid.min())
            xlabel = f'x (${length_unit}$)'
            ylabel = f'y (${length_unit}$)'
        
        # Set default figsize if needed
        if figsize is None:
            xscale = 1.175
            if yscale < 1:
                figsize = (xscale * 10, 10 * yscale)
            else:
                figsize = (xscale * 10 / yscale, 10)
        
        # Generate plot
        fig = plt.figure(figsize=figsize)
        plt.pcolormesh(x_grid, y_grid, C, **kwargs)
        plt.xlabel(xlabel, fontsize='x-large')
        plt.ylabel(ylabel, fontsize='x-large')
        cbar = plt.colorbar(aspect=40, fraction=0.1)
        cbar.ax.set_ylabel(f'$δ_{{gsf}}$ (${length_unit}$)', fontsize='x-large')
        
        return fig
    
    def delta_line_plot(self, vect=None, num=None, smooth=True,
                        length_unit='Å',
                        figsize=None, fig=None, **kwargs):
        """
        Generates a line plot for the interpolated delta planar shift values
        along a specified crystallographic vector in the (a1, a2) plane.
        
        Parameters
        ----------
        vect : numpy.array, optional
            Vector to plot the gsf along.  If box is set, this vect will be a
            lattice vector, otherwise it will be a Cartesian vector.  Must be 
            in the plane defined by the GammaSurface object's a1vect and 
            a2vect vectors.  Default value will use the set a1vect.
        num : int, optional
            The number of points to evaluate the generalized stacking fault
            energy for.  Default value is 100 if smooth is True, otherwise is
            number of unique a1 values from 0 to 1.
        smooth : bool, optional
            If True (default), then plot shows smooth interpolated values.
            If False, plot shows nearest raw data values.
        length_unit : str, optional
            The unit of length to display the x-axis coordinates in.
            Default value is 'Å'.
        figsize : tuple, optional
            The x,y size of the figure to return.  Default value is (10, 6).
        fig : matplotlib.figure, optional
            An existing figure object to add the new plot to.  If not given, a
            new figure is generated.
        **kwargs : dict, optional
            Additional keywords are passed into the underlying 
            matplotlib.pyplot.plot(). This allows control of such things
            like line color, style, etc.
        
        Returns
        -------
        matplotlib.figure
        """
        if not self.__hasdata:
            raise AttributeError('gamma surface data not set')
        if 'delta' not in self.data:
            raise AttributeError('delta data not set')

        if num is None:
            if smooth:
                num = 100
            else:
                unique_a1 = np.unique([self.data.a1-1, self.data.a1, self.data.a1+1])
                num = len(unique_a1[(unique_a1 >=-0.000001) & (unique_a1 <=1.000001)])

        # Generate coordinates
        a1 = np.linspace(0, 1, num)
        a2 = np.zeros(num)
        pos = self.a12_to_pos(a1, a2, a1vect=vect)
        
        # Evaluate interpolated distance along x
        d = uc.get_in_units(self.delta(pos=pos, smooth=smooth), length_unit)
        x = uc.get_in_units(np.linalg.norm(pos, axis=1), length_unit)
        
        # Create plot
        xmax = x.max()
        dmin = d.min() * 1.05
        dmax = d.max() * 1.05
        if fig is None:
            if figsize is None:
                figsize = (10, 6)
            fig = plt.figure(figsize=figsize)
        else:
            old_xmax = fig.axes[0].get_xlim()[-1]
            old_dmin = fig.axes[0].get_ylim()[0]
            old_dmax = fig.axes[0].get_ylim()[-1]
            if old_xmax > xmax:
                xmax = old_xmax
            if old_dmin < dmin:
                dmin = old_dmin
            if old_dmax > dmax:
                dmax = old_dmax
        if 'fmt' in kwargs:
            fmt = kwargs.pop('fmt')
            plt.plot(x, d, fmt, **kwargs)
        else:
            plt.plot(x, d, **kwargs)
        
        if vect is None:
            vect = self.a1vect

        plt.xlabel(f'$x$ along {vect} (${length_unit}$)', fontsize='x-large')
        plt.ylabel(f'$δ_{{gsf}}$ (${length_unit}$)', fontsize='x-large')
        plt.xlim(0, xmax)
        plt.ylim(dmin, dmax)
        
        return fig

    def path(self, coord, style='ISM', gradientfxn='cdiff',
             gradientkwargs=None, integratorfxn='rk'):
        """
        Creates an mep Path object mapping for the gamma surface based on
        supplied xy coordinates along the path line.
        
        Parameters
        ----------
        coord : array-like object
            The xy coordinates of the points along the path.
        style : str
            The path/relaxer style to use.  Default value of 'ISM' will use improved string method.
        gradientfxn : function, optional
            The function to use to estimate the gradient of the energy.  Default
            value of 'cdiff' will use atomman.mep.gradient.central_difference
        gradientkwargs : dict or None, optional
            The keyword arguments (i.e. settings) to use with the gradientfxn.
            Default value of None will use {'shift':1e-7}.
        integratorfxn : str or function, optional
            The function to use to integrate relaxation steps.  Default value of
            'rk' will use atomman.mep.integrator.rungekutta.
            
        Returns
        -------
        subclass of atomman.mep.BasePath
            Specific class dictated by style: style=='ISM' -> ISMPath (only style currently).    
        """
        # Handle default values
        if gradientkwargs is None:
            gradientkwargs = {'shift':1e-7}
        
        # Define energyfxn for the path
        def energyfxn(xy):
            return self.E_gsf(x=xy[..., 0], y=xy[..., 1])

        return create_path(coord, energyfxn, style=style, gradientfxn=gradientfxn,
                           gradientkwargs=gradientkwargs, integratorfxn=integratorfxn)

    def build_path(self, pos, npoints=31, style='ISM', gradientfxn='cdiff',
                   gradientkwargs=None, integratorfxn='rk'):
        """
        Builds a subclass of atomman.mep.BasePath as one or two line segments
        along a gamma surface. The energy function along the path will be
        properly set using E_gsf.
        
        Parameters
        ----------
        pos : array-like object
            2x3 or 3x3 array of Miller vector points that defines the end points of the path's
            line segment(s).
        npoints : int, optional
            The number of points to include along the path. Must be odd if three pos are used.
            Default value is 31.
        style : str
            The path/relaxer style to use.  Default value of 'ISM' will use improved string method.
        gradientfxn : function, optional
            The function to use to estimate the gradient of the energy.  Default
            value of 'cdiff' will use atomman.mep.gradient.central_difference
        gradientkwargs : dict or None, optional
            The keyword arguments (i.e. settings) to use with the gradientfxn.
            Default value of None will use {'shift':1e-7}.
        integratorfxn : str or function, optional
            The function to use to integrate relaxation steps.  Default value of
            'rk' will use atomman.mep.integrator.rungekutta.
            
        Returns
        -------
        subclass of atomman.mep.BasePath
            Specific class dictated by style: style=='ISM' -> ISMPath (only style currently).    
        """
        
        # Define path coordinate builder
        def build_coord(startpos, endpos, npoints):

            box = self.box
            x, y = self.pos_to_xy(miller.vector_crystal_to_cartesian(startpos, box))
            start_xy = np.array([x, y])

            x, y = self.pos_to_xy(miller.vector_crystal_to_cartesian(endpos, box))
            end_xy = np.array([x, y])

            return np.vstack([np.linspace(start_xy[0], end_xy[0], npoints),
                              np.linspace(start_xy[1], end_xy[1], npoints)]).T
        
        # Construct single segment path coordinates
        if len(pos) == 2:
            coord = build_coord(pos[0], pos[1], npoints)
        
        # Construct double segment path coordinates
        elif len(pos) == 3:
            assert npoints % 2 == 1, 'npoints must be odd'
            npoints_2 = int(npoints + 1 / 2)
            coorda = build_coord(pos[0], pos[1], npoints_2)
            coordb = build_coord(pos[1], pos[2], npoints_2)
            coord = np.vstack([coorda, coordb[1:]])
            
        else:
            raise ValueError('pos must have 2 or 3 coordinates')
        
        return self.path(coord, style=style, gradientfxn=gradientfxn,
                        gradientkwargs=gradientkwargs, integratorfxn=integratorfxn)
    


# coding: utf-8
# Standard Python libraries
import os

def dump(system, f=None, box_unit=None, prop_name=None, unit=None,
         prop_unit=None, format=None, indent=None):
    """
    Dumps a JSON/XML System.model() representation of the system.
    
    Parameters
    ----------
    system : atomman.System
        The system to generate the data model for.
    f : str or file-like object, optional
        File path or file-like object to write the content to.  If not given,
        then the content is returned as a DataModelDict.
    box_unit : str, optional
        Length unit to use for the box. Default value is 'angstrom'.
    prop_name : list, optional
        The Atoms properties to include.  If neither prop_name nor prop_unit
        are given, all system properties will be included.
    unit : list, optional
        Lists the units for each prop_name as stored in the table.  For a
        value of None, no conversion will be performed for that property.  For
        a value of 'scaled', the corresponding table values will be taken in
        box-scaled units.  If neither unit nor prop_units given, pos will be
        given in Angstroms and all other values will not be converted.
    prop_unit : dict, optional
        dictionary where the keys are the property keys to include, and
        the values are units to use. If neither unit nor prop_units given, 
        pos will be given in Angstroms and all other values will not be
        converted.
    format : str, optional
        File format 'xml' or 'json' to save the content as if f is given.  If
        f is a filename, then the format will be automatically inferred from
        f's extension.  If format is not given and cannot be inferred, then it
        will be set to 'json'.
    indent : int or None, optional
        Indentation option to use for XML/JSON content if f is given.  A value
        of None (default) will add no line separatations or indentations.

    Returns
    -------
    model : DataModelDict.DataModelDict or str
        The generated model representation of the system.  Will be a
        DataModelDict if format is not specified, and a JSON- or XML-formatted
        string if format is specified.  Returned if f is not given.
    """
        
    # Generate model for system
    model = system.model(box_unit=box_unit, prop_name=prop_name, unit=unit,
                         prop_unit=prop_unit)

    # Return DataModelDict or str
    if f is None:
        if format is None:
            return model
        elif format.lower() == 'xml':
            return model.xml(indent=indent)
        elif format.lower() == 'json':
            return model.json(indent=indent)
    
    # Write to file
    else:
        if format is None:
            try:
                format = os.path.splitext(f)[1][1:]
            except:
                format = 'json'
        
        if hasattr(f, 'write'):
            if format.lower() == 'xml':
                return model.xml(fp=f, indent=indent)
            elif format.lower() == 'json':
                return model.json(fp=f, indent=indent)
        
        else:
            with open(f, 'w') as fp:
                if format.lower() == 'xml':
                    return model.xml(fp=fp, indent=indent)
                elif format.lower() == 'json':
                    return model.json(fp=fp, indent=indent)


import numpy as np

from .. import Box
from ..tools import miller, vect_angle

def dislocation_system_transform(ξ_uvw, slip_hkl, m=None, n=None, box=None, tol=1e-8):
    """
    Utility function for getting the transformation matrix to the dislocation system
    based on crystal orientation vectors.
    
    Parameters
    ----------
    ξ_uvw : array-like object
        The Miller crystal vector associated with the dislocation's line direction.
    slip_hkl : array-like object
        The Miller plane indices associated with the dislocation's slip plane.
    m : array-like object, optional
        The m unit vector for the solution.  m, n, and u (dislocation line)
        should be right-hand orthogonal.  Default value is [1, 0, 0]
        (x-axis).
    n : array-like object, optional
        The n unit vector for the solution.  m, n, and u (dislocation line)
        should be right-hand orthogonal.  Default value is [0, 1, 0]
        (y-axis).
    box : atomman.Box, optional
        The unit cell's box that the crystal vectors are taken with respect to. If
        not given, will use a cubic box with a=1 (vects are taken as Cartesian).
    tol : float, optional
        Tolerance parameter used to round off near-zero values.  Default
        value is 1e-8.
    """
    # Set default box to be cubic
    if box is None:
        box = Box()

    # Check m value
    if m is None:
        m = np.array([1,0,0], dtype=float)
    m = np.asarray(m, dtype=float)
    assert m.shape == (3, ), "m must be a 3D vector"
    assert np.isclose(np.linalg.norm(m), 1.0, atol = tol), "m must be a unit vector"

    # Check n value
    if n is None:
        n = np.array([0,1,0], dtype=float)
    n = np.asarray(n, dtype=float)
    assert n.shape == (3, ), "n must be a 3D vector"
    assert np.isclose(np.linalg.norm(n), 1.0, atol = tol), "n must be a unit vector"
    assert np.isclose(np.dot(m, n), 0.0, atol = tol), "m and n must be perpendicular"
    
    # ξ_axis is the Cartesian unit vector along dislocation line 
    ξ_axis = miller.vector_crystal_to_cartesian(ξ_uvw, box)
    ξ_axis = ξ_axis / np.linalg.norm(ξ_axis)
    
    # n_axis is the unit vector normal to slip plane
    n_axis = miller.plane_crystal_to_cartesian(slip_hkl, box)
    
    # m_axis is the unit vector in the "edge" direction
    m_axis = np.cross(n_axis, ξ_axis)
    
    transform = np.array([m_axis, n_axis, ξ_axis])
    
    # Transform transform to m,n,ξ orientation
    T = np.array([m,n,np.cross(m, n)]).T
    
    transform = T.dot(transform)
    
    return transform


# coding: utf-8

# https://pandas.pydata.org/
import pandas as pd

# atomman imports
import atomman.unitconvert as uc
from .atoms_prop_info import atoms_prop_info
from .velocities_prop_info import velocities_prop_info
from ... import Atoms, Box, System
from ...lammps import style
from .. import load_table, FileFormatError
from ...tools import uber_open_rmode

def load(data, pbc=(True, True, True), symbols=None, atom_style=None,
         units='metal'):
    """
    Read a LAMMPS-style atom data file.
    
    Parameters
    ----------
    data : str or file-like object
        The atom data content to read.  Can be str content, path name, or open
        file-like object.
    pbc : list of bool
        Three boolean values indicating which System directions are periodic.
        Default value is (True, True, True).
    symbols : tuple, optional
        Allows the list of element symbols to be assigned during loading.
    atom_style : str, optional
        The LAMMPS atom_style option associated with the data file.  Optional
        as the data can list this value in a comment in the Atoms section
        header.  If not given and not found in data, the default value of
        'atomic' is used.
    units : str, optional
        The LAMMPS units option associated with the data file.  Default value
        is 'metal'.
    
    Returns
    -------
    atomman.System
        The corresponding system.  Note all property values will be
        automatically converted to atomman.unitconvert's working units.

    Raises
    ------
    FileNotFoundError
        If data is (likely) a file name and no matching file is found.
    FileFormatError
        If required content fields not found.
    ValueError
        If atom_style is both given as a parameter and found in data, but are
        not the same
    """

    # First pass over file to generate system and locate content
    system, params = firstpass(data, pbc, symbols, units)

    # Set atom_style
    if atom_style is None:
        
        # Use default if no value given/found
        if params['atom_style'] is None:
            atom_style = 'atomic'
        
        # Use value listed in data
        else:
            atom_style = params['atom_style']
    
    # Check that atom_style matches the one in data
    elif params['atom_style'] is not None and atom_style != params['atom_style']:
        raise ValueError(f'given atom_style of {atom_style} differs from value of {params["atom_style"]} found in data')
    
    # Read in Atoms info
    system = read_atoms(data, system, atom_style, units, params['atomsstart'], params['atomscolumns'])
    
    # Read in Velocities info
    system = read_velocities(data, system, atom_style, units, params['velocitiesstart'])
    
    return system

def firstpass(data, pbc, symbols, units):
    """
    Reads through data to extract natoms and box dimensions to construct a
    System sans atomic properties.

    Parameters
    ----------
    data : str or file-like object
        The atom data content to read.  Can be str content, path name, or open
        file-like object.
    pbc : list of bool
        Three boolean values indicating which System directions are periodic.
        Default value is (True, True, True).
    symbols : tuple, optional
        Allows the list of element symbols to be assigned during loading.
    units : str 
        The LAMMPS units option associated with the data file. Default value
        is 'metal'.
    
    Returns
    -------
    system : atomman.System
        The system with the correct number of atoms and no assigned atom
        properties.
    params : dict
        Line indices and flags indicating how to read in the atom properties
        from the file on subsequent passes.

    Raises
    ------
    FileFormatError
        If '# atoms' or box boundaries not found.
    """
    # Get units information
    units_dict = style.unit(units)
    
    # Initialize parameter values
    atomsstart = None
    velocitiesstart = None
    natoms = None
    natypes = None
    firstatoms = False
    atomscolumns = 0
    masses = None
    num_masses_to_read = 0
    xlo = xhi = ylo = yhi = zlo = zhi = None
    xy = 0.0
    xz = 0.0
    yz = 0.0
    i = 0
    
    # Read str and files in the same way
    with uber_open_rmode(data) as fp:
        
        # Loop over all lines in fp
        for i, fullline in enumerate(fp):
            try:
                fullline = fullline.decode('UTF-8')
            except:
                pass
            
            # Remove comments after '#'
            try:
                comment_index = fullline.index('#')
            except:
                line = fullline
            else:
                line = fullline[:comment_index]
            
            terms = line.split()

            # Skip blank lines
            if len(terms)>0:
                
                # Read number of atoms 
                if len(terms) == 2 and terms[1] == 'atoms':
                    natoms = int(terms[0])

                # Read number of atom types
                elif len(terms) == 3 and terms[1] == 'atom' and terms[2] == 'types': 
                    natypes = int(terms[0])
                
                # Read boundary info
                elif len(terms) == 4 and terms[2] == 'xlo' and terms[3] == 'xhi':
                    xlo = uc.set_in_units(float(terms[0]), units_dict['length'])
                    xhi = uc.set_in_units(float(terms[1]), units_dict['length'])

                elif len(terms) == 4 and terms[2] == 'ylo' and terms[3] == 'yhi':
                    ylo = uc.set_in_units(float(terms[0]), units_dict['length'])
                    yhi = uc.set_in_units(float(terms[1]), units_dict['length'])

                elif len(terms) == 4 and terms[2] == 'zlo' and terms[3] == 'zhi':
                    zlo = uc.set_in_units(float(terms[0]), units_dict['length'])
                    zhi = uc.set_in_units(float(terms[1]), units_dict['length'])

                elif len(terms) == 6 and terms[3] == 'xy' and terms[4] == 'xz' and terms[5] == 'yz':
                    xy = uc.set_in_units(float(terms[0]), units_dict['length'])
                    xz = uc.set_in_units(float(terms[1]), units_dict['length'])
                    yz = uc.set_in_units(float(terms[2]), units_dict['length'])
                
                # Identify starting line number for Atoms data
                elif len(terms) == 1 and terms[0] == 'Atoms':
                    atomsstart = i + 1
                    firstatoms = True

                    # Check for atom_style comment
                    try: 
                        comment_index = fullline.index('#')
                    except:
                        atom_style = None
                    else:
                        atom_style = fullline[comment_index + 1:].strip()
                
                # Count number of columns in Atoms table
                elif firstatoms:
                    atomscolumns = len(terms)
                    firstatoms = False
                
                # Identify starting line for Masses data
                elif len(terms) == 1 and terms[0] == 'Masses':
                    if natypes is None:
                        raise FileFormatError('# atom types must appear before Masses list')
                    masses = [None for i in range(natypes)]
                    num_masses_to_read = natypes
                
                # Read masses
                elif num_masses_to_read > 0:
                    read_mass(terms, masses)
                    num_masses_to_read -= 1

                # Identify starting line number for Velocity data
                elif len(terms) == 1 and terms[0] == 'Velocities':
                    velocitiesstart = i + 1
    
    if i == 0:
        raise FileNotFoundError(f'File {data} not found')

    if natoms is None:
        raise FileFormatError('# atoms not found')

    if xlo is None or xhi is None:
        raise FileFormatError('xlo, xhi box dimensions missing')

    if ylo is None or yhi is None:
        raise FileFormatError('ylo, yhi box dimensions missing')

    if zlo is None or zhi is None:
        raise FileFormatError('zlo, zhi box dimensions missing')

    if atomsstart is None:
        raise FileFormatError('Atoms section missing')

    # Create system with natoms
    box = Box(xlo=xlo, xhi=xhi,
              ylo=ylo, yhi=yhi,
              zlo=zlo, zhi=zhi,
              xy=xy, xz=xz, yz=yz)
    atoms = Atoms(natoms=natoms)
    system = System(box=box, atoms=atoms, pbc=pbc, symbols=symbols,
                    masses=masses)

    # Compile dict of params
    params = {}
    params['atomsstart'] = atomsstart
    params['velocitiesstart'] = velocitiesstart
    params['atomscolumns'] = atomscolumns
    params['atom_style'] = atom_style

    return system, params

def read_atoms(data, system, atom_style, units, atomsstart, atomscolumns):
    """
    Reads in an "Atoms" table from data.

    Parameters
    ----------
    data : str or file-like object
        The atom data content to read.  Can be str content, path name, or open
        file-like object.
    system : atomman.System
        The system to add Atoms table info to.
    atom_style :str
        The LAMMPS atom_style option associated with the data file.  Default
        value is 'atomic'.
    units : str 
        The LAMMPS units option associated with the data file. Default value
        is 'metal'.
    atomsstart : int or None
        The line of the file where the Atoms table content starts.
    atomscolumns : int
        How many columns are in the Atoms table.  Used to determine if wrap
        flags are included.

    Returns
    -------
    atomman.System
        The system with atom properties from the Atoms table assigned.
    """
    
    if atomsstart is not None:
        prop_info = atoms_prop_info(atom_style, units)
        ncols = countreadcolumns(prop_info)
        
        # Read Atoms table
        system = load_table(data, box=system.box, system=system, 
                            prop_info=prop_info, skiprows=atomsstart,
                            nrows=system.natoms, comment='#',
                            header=None, usecols=range(ncols))
        
        # Check if image flags are included
        if atomscolumns == ncols + 3:
            
            # Read image flags
            with uber_open_rmode(data) as f:
                imageflags = pd.read_csv(f, delim_whitespace=True, names=['bx', 'by', 'bz'],
                                        skiprows=atomsstart, nrows=system.natoms, comment='#',
                                        header=None, usecols=range(ncols, atomscolumns),
                                        dtype='int64')

            # Wrap atoms to correct images
            shift = imageflags.values.dot(system.box.vects)
            system.atoms.pos[:] += shift
        
        # Check for correct number of columns
        elif ncols != atomscolumns:
            raise FileFormatError(f'atom_style={atom_style} requires {ncols} or {ncols+3} Atoms table columns but {atomscolumns} found')

    return system

def countreadcolumns(prop_info):
        """
        Counts how many columns are expected from prop info
        """
        count = 0
        for prop in prop_info:
            if isinstance(prop['table_name'], str):
                count += 1
            else:
                count += len(prop['table_name'])
        return count

def read_mass(terms, masses):
    """
    Reads in a line in the Masses table.

    Parameters
    ----------
    terms : list
        The space-delimited terms in an input line
    masses : list
        The list of already assigned masses
    """
    # Check that Masses line is correctly formatted
    try:
        assert len(terms) == 2
        atype = int(terms[0])
        assert atype > 0 and atype <= len(masses)
        mass = float(terms[1])
        assert mass > 0
    except:
        raise FileFormatError('Invalid mass term')
    
    if masses[atype - 1] is None:
        masses[atype - 1] = mass
    else:
        raise FileFormatError(f'Multiple masses listed for atom type {atype}')

def read_velocities(data, system, atom_style, units, velocitiesstart):
    """
    Reads in an "Velocities" table from data.

    Parameters
    ----------
    data : str or file-like object
        The atom data content to read.  Can be str content, path name, or open
        file-like object.
    system : atomman.System
        The system to add Velocities table info to.
    atom_style :str
        The LAMMPS atom_style option associated with the data file.  Default
        value is 'atomic'.
    units : str 
        The LAMMPS units option associated with the data file. Default value
        is 'metal'.
    velocitiesstart : int or None
        The line of the file where the Velocities table content starts.

    Returns
    -------
    atomman.System
        The system with atom properties from the Velocities table assigned.
    """
    if velocitiesstart is not None:
        prop_info = velocities_prop_info(atom_style, units)
        system = load_table(data, box=system.box, system=system,
                            prop_info=prop_info, skiprows=velocitiesstart,
                            nrows=system.natoms, comment='#', header=None)


    return system



# coding: utf-8
# Standard Python libraries
from copy import deepcopy

# http://www.numpy.org/
import numpy as np

# atomman imports
from . import VolterraDislocation
from ..tools import axes_check

class Stroh(VolterraDislocation):
    """
    Class for solving the Eshelby anisotropic solution for a straight
    dislocation or crack using the Stroh method.
    """
    
    def solve(self, C, burgers, ξ_uvw=None, slip_hkl=None, transform=None,
              axes=None, box=None, m=[1,0,0], n=[0,1,0], tol=1e-8):
        """
        Computes the elastic solution for an anisotropic volterra dislocation.
        
        Parameters
        ----------
        C : atomman.ElasticConstants
            The medium's elastic constants.
        burgers : array-like object
            The dislocation's Burgers vector.
        ξ_uvw : array-like object
            The Miller crystal vector associated with the dislocation's line
            direction.  Must be given with slip_hkl to identify the
            transformation matrix to use on C and burgers.
        slip_hkl : array-like object
            The Miller plane indices associated with the dislocation's slip
            plane.  Must be given with slip_hkl to identify the
            transformation matrix to use on C and burgers.
        transform : array-like object, optional
            A 3x3 set of orthogonal Cartesian vectors that define the
            transformation matrix to use on C and burgers to convert from the
            standard (unit cell) and dislocation orientations.  The 3 vectors
            will automatically be converted into unit vectors.  Using this is
            an alternative to using ξ_uvw and slip_hkl.
        axes : array-like object, optional
            Same as transform.  Retained for backwards compatibility.
        box : atomman.Box, optional
            The unit cell's box that crystal vectors are taken with respect to.
            If not given, will use a cubic box with a=1 meaning that burgers,
            ξ_uvw and slip_hkl will be interpreted as Cartesian vectors.
        m : array-like object, optional
            The m unit vector for the solution.  m, n, and u (dislocation
            line) should be right-hand orthogonal.  Default value is [1,0,0]
            (x-axis).
        n : array-like object, optional
            The n unit vector for the solution.  m, n, and u (dislocation
            line) should be right-hand orthogonal.  Default value is [0,1,0]
            (y-axis). n is normal to the dislocation slip plane.
        tol : float
            Tolerance parameter used to round off near-zero values.  Default
            value is 1e-8.
        """
        VolterraDislocation.solve(self, C, burgers, ξ_uvw=ξ_uvw, slip_hkl=slip_hkl,
                                  transform=transform, axes=axes, box=box,
                                  m=m, n=n, tol=tol)

        # Pull out full 3x3x3x3 elastic constants matrix
        Cijkl = self.C.Cijkl
        
        # Matrices of Cijkl constants used to construct N
        mm = np.einsum('i,ijkl,l', self.m, Cijkl, self.m)
        mn = np.einsum('i,ijkl,l', self.m, Cijkl, self.n)
        nm = np.einsum('i,ijkl,l', self.n, Cijkl, self.m)
        nn = np.einsum('i,ijkl,l', self.n, Cijkl, self.n)
        
        # The four 3x3 matrices that represent the quadrants of N
        NB = -np.linalg.inv(nn)
        NA = NB.dot(nm)
        NC = mn.dot(NA) + mm
        ND = mn.dot(NB)
        
        # N is the 6x6 array, where the eigenvalues are the roots p
        # and the eigenvectors give A and L
        N =  np.array(np.vstack((np.hstack((NA, NB)), np.hstack((NC, ND)))))
        
        # Calculate the eigenvectors and eigenvalues
        eig = np.linalg.eig(N)
        p = eig[0]
        eigvec = np.transpose(eig[1])
        
        # Separate the eigenvectors into A and L
        A = np.array([eigvec[0,:3], eigvec[1,:3], eigvec[2,:3], eigvec[3,:3], eigvec[4,:3], eigvec[5,:3]])
        L = np.array([eigvec[0,3:], eigvec[1,3:], eigvec[2,3:], eigvec[3,3:], eigvec[4,3:], eigvec[5,3:]])
        
        # Calculate normalization factor k
        k = 1. / (2. * np.einsum('si,si->s', A, L))
        
        # Calculation verification checks
        try:
            assert np.allclose(np.einsum('s,si,sj->ij', k,A,L), np.identity(3, dtype='complex128'), atol=tol)
            assert np.allclose(np.einsum('s,si,sj->ij', k,A,A), np.zeros((3,3), dtype='complex128'), atol=tol)
            assert np.allclose(np.einsum('s,si,sj->ij', k,L,L), np.zeros((3,3), dtype='complex128'), atol=tol)
            assert np.allclose(np.einsum('s,t,si,ti->st', k**.5,k**.5,A,L) + np.einsum('s,t,ti,si->st', k**.5,k**.5,A,L),
                                         np.identity(6, dtype='complex128'), atol = tol)
        except:
            raise ValueError('Stroh checks failed!')
        
        # Assign property values
        self.__Cijkl = Cijkl
        self.__p = p
        self.__A = A
        self.__L = L
        self.__k = k
    
        # Check that K_tensor is real
        if self.K_tensor.dtype == 'complex128':
            raise ValueError('Solution not real: check elastic constants')

    @property
    def p(self):
        """numpy.ndarray : p eigenvalues"""
        return deepcopy(self.__p)
    
    @property
    def A(self):
        """numpy.ndarray : A eigenvectors"""
        return deepcopy(self.__A)
    
    @property
    def L(self):
        """numpy.ndarray : L eigenvectors"""
        return deepcopy(self.__L)
    
    @property
    def k(self):
        """numpy.ndarray : k normalization factors"""
        return deepcopy(self.__k)
    
    @property
    def K_coeff(self):
        """float : The energy coefficient"""
        
        # K = b_i K_ij b_j / (b_k b_k)
        return (self.burgers.dot(self.K_tensor.dot(self.burgers))
                / self.burgers.dot(self.burgers))
    
    @property
    def K_tensor(self):
        """numpy.ndarray : The anisotropic energy coefficient tensor"""
        
        # ii is imaginary unit
        ii = np.array([1.j])
        
        # updn is alternating +-
        updn = np.array([1, -1, 1, -1, 1, -1])
        
        # Compute K_tensor
        K = ii * np.einsum('s,s,si,sj->ij', updn, self.k, self.L, self.L)
        
        # Round away near-zero terms
        K = np.real_if_close(K, tol=self.tol)
        K[np.isclose(K / K.max(), 0.0, atol=self.tol)] = 0.0
        
        return K
    
    @property
    def preln(self):
        """float : The pre-ln strain energy factor"""
        # a = b_i K_ij b_j / (4 π)
        return self.burgers.dot(self.K_tensor.dot(self.burgers)) / (4 * np.pi)
    
    def displacement(self, pos):
        """
        Compute the position-dependent anisotropic displacement.
        
            u_i = 1 / (2 π i) (Σ_a +- k_a A_ai (L_aj*burgers_j) ln(η_a))
        
        Parameters
        ----------
        pos : array-like object
            3D vector position(s).
        
        Returns
        -------
        numpy.ndarray
            The computed 3D vector displacements at all given points.
        """
        # ii is imaginary unit
        ii = np.array([1.j])
        
        # updn is alternating +-
        updn = np.array([1, -1, 1, -1, 1, -1])
        
        # Convert pos to eta coordinate
        eta = self.eta(pos)
        
        # Compute the displacements
        disp = 1 / (2 * np.pi * ii) * np.einsum('a,a,ai,a,na->ni',
                                                updn, self.k, self.A,
                                                self.L.dot(self.burgers),
                                                np.log(eta))
        
        # Round away near-zero terms
        disp = np.real_if_close(disp, tol=self.tol)
        
        # Reduce single-value solutions
        if np.asarray(pos).ndim == 1:
            return disp[0]
        else:
            return disp
    
    def stress(self, pos):
        """
        Compute the position-dependent anisotropic stresses.
        
            σ_ij = 1 / (2 π i) (Σ_a +- k_a C_ijkl mpn_al A_ak (L_am*burgers_m) / η_a)
        
        Parameters
        ----------
        pos : array-like object
            3D vector position(s).
        
        Returns
        -------
        numpy.ndarray
            The computed 3x3 stress states at all given points.
        """
        # ii is imaginary unit
        ii = np.array([1.j])
        
        # updn is alternating +-
        updn = np.array([1, -1, 1, -1, 1, -1])
        
        # Convert pos to eta coordinate
        eta = self.eta(pos)
        
        # Compute mpn factor
        mpn = self.m + np.outer(self.p, self.n)
        
        # Compute the stresses
        stress = 1 / (2 * np.pi * ii) * np.einsum('a,a,ijkl,al,ak,a,na->nij',
                                                  updn, self.k, self.C.Cijkl,
                                                  mpn, self.A,
                                                  self.L.dot(self.burgers),
                                                  1/eta)
        
        # Round away near-zero terms
        stress = np.real_if_close(stress, tol=self.tol)
        
        if np.asarray(pos).ndim == 1:
            return stress[0]
        else:
            return stress
    
    def eta(self, pos):
        """
        Compute the eta coordinates based on positions, p, m and n.  Used by
        displacement() and stress().
        
            η_a = x_i m_i + p_a x_j n_j
        
        Parameters
        ----------
        pos : array-like object
            3D vector position(s).
        
        Returns
        -------
        numpy.ndarray
            The computed eta factor at all given points.
        """
        
        x = np.dot(pos, self.m)
        y = np.dot(pos, self.n)
        
        return (x + np.outer(self.p, y)).T


# coding: utf-8

def filltemplate(template, variable, s_delimiter, e_delimiter):
    """
    Takes a template and fills in values for delimited template variables.
    
    Parameters
    ----------
    template : string or file-like object
        The template file or file content to fill in.
    variable : dict 
        Dictionary with keys defining the delimited template variable terms,
        and values the values to replace the variable terms with.
    s_delimiter : str
        The leading delimiter for identifying the template variable terms.
    e_delimiter : str 
        The trailing delimiter for identifying the template variable terms.
    
    Returns
    -------
    str
        The template with all delimited variable terms replaced with their
        corresponding defined values from variable.
        
    Raises
    ------
    KeyError
        If delimited term found in template that has no value in variable.
    ValueError
        If parsing of s_delimiter, e_delimiter pairs fails.
    """
    
    # Convert to string if a file-like object
    try:
        template = template.read()
    except AttributeError:
        pass
    
    # Loop until done
    while True:
        
        # Search for starting delimiter
        try:
            s = template.index(s_delimiter)
        except ValueError: 
            s = None
        else:
            s = s + len(s_delimiter)
        
        # Search for ending delimiter
        try:
            e = template.index(e_delimiter)
        except ValueError:
            e = None        
        
        # Replace delimited string with value
        if s is not None and e is not None and s < e:
            name = template[s: e]
            var = s_delimiter + name + e_delimiter
            try:
                value = str(variable[name])
            except KeyError:
                raise KeyError(name + ' not found in variable dictionary')
            template = template.replace(var, value)
        
        # Finish if no delimiters remain
        elif s is None and e is None:
            break
            
        # Issue errors
        elif s is None:
            raise ValueError('ending delimiter found without starting delimiter')
        elif e is None:
            raise ValueError('starting delimiter found without ending delimiter')
        else:
            raise ValueError('ending delimiter found before starting delimiter')
            
    return template


# coding: utf-8
# Standard Python libraries
from copy import deepcopy

# atomman imports
from ...tools import indexstr

def process_prop_info(prop_name=None, table_name=None, shape=None, unit=None, dtype=None, prop_info=None):
    """Handles common setting of prop_info terms"""
    
    if prop_info is not None:
        # Check that competing parameters are not given
        try:
            assert prop_name is None
            assert table_name is None
            assert shape is None
            assert unit is None
            assert dtype is None
        except:
            raise ValueError('prop_info cannot be given with prop_name, table_name, shape, unit, or dtype')
        prop_info = deepcopy(prop_info)
    
    # Combine list parameters into prop_info
    else:
        
        # Check that prop_name is given
        if prop_name is None:
            raise ValueError('prop_info or prop_name is required.')
        numprops = len(prop_name)
        
        # Check that sizes match
        try:
            assert table_name is None or len(table_name) == numprops
            assert shape is None or len(shape) == numprops
            assert unit is None or len(unit) == numprops
            assert dtype is None or len(dtype) == numprops
        except:
            raise ValueError('any of prop_name, table_name, shape, unit, and dtype given must be the same length')
        
        # Build prop_info
        prop_info = []
        for i in range(numprops):
            pinfo = {}
            pinfo['prop_name'] = prop_name[i]
            if table_name is not None:
                pinfo['table_name'] = table_name[i]
            if shape is not None:
                pinfo['shape'] = shape[i]
            if unit is not None:
                pinfo['unit'] = unit[i]
            if dtype is not None:
                pinfo['dtype'] = dtype[i]
            prop_info.append(pinfo)
    
    # Fill in missing default values
    for prop in prop_info:
        
        # prop_name is required
        if 'prop_name' not in prop:
            raise KeyError('prop_name required for each entry in prop_info')
        
        # Generate table_name from prop_name and shape
        if 'table_name' not in prop:
            prop['shape'] = prop.get('shape', ())
            prop['table_name'] = []
            for index, istr in indexstr(prop['shape']): # pylint: disable=unused-variable
                prop['table_name'].append(prop['prop_name'] + istr)
        # Make certain table_name is a list
        else:
            try:
                assert not isinstance(prop['table_name'], str)
                prop['table_name'] = list(prop['table_name'])
            except:
                prop['table_name'] = [prop['table_name']]
        
        # Take shape as length of table_name
        if 'shape' not in prop:
            numtnames = len(prop['table_name'])
            if numtnames == 1:
                prop['shape'] = ()
            else:
                prop['shape'] = (numtnames, )
        # Make certain shape is a tuple
        else:
            try:
                prop['shape'] = tuple(prop['shape'])
            except:
                prop['shape'] = (prop['shape'],)
        
        # Default unit and dtype values are None
        prop['unit'] = prop.get('unit', None)
        prop['dtype'] = prop.get('dtype', None)
    
    return prop_info


# coding: utf-8
# Standard Python libraries
import os
import shutil
import tempfile

# atomman imports
from . import run

def checkversion(lammps_command):
    """
    Gets the LAMMPS version and date information by testing lammps_command.
    
    Parameters
    ----------
    lammps_command: str
        A LAMMPS executable.
        
    Returns
    -------
    version_info: dict
        Dictionary containing 'version', the str LAMMPS version, and
        'date', the corresponding  datetime.date for the LAMMPS 
        version.
        
    Raises
    ------
    ValueError
        If lammps fails to run
    """
    # Define emptyscript and logfile paths
    tempdir = tempfile.mkdtemp()
    emptyscript = os.path.join(tempdir, 'empty.in')
    logfile = os.path.join(tempdir, 'empty.lammps')
    
    # Create emptyscript
    with open(emptyscript, 'w') as f:
        f.write('')
    
    # Run lammps_command with emptyscript
    try:
        log = run(lammps_command, emptyscript, logfile=logfile)
    except:
        raise ValueError('Failed to run simulation with lammps_command '+lammps_command)
    finally:
        shutil.rmtree(tempdir)
    
    # Extract lammps version and date info
    version_info = {}
    version_info['version'] = log.lammps_version
    version_info['date'] = log.lammps_date
    return version_info